import json

import boto3
import pytest

from moto import mock_aws

template = """{
  "AWSTemplateFormatVersion": "2010-09-09",
  "Description": "The AWS CloudFormation template for this Serverless application",
  "Resources": {
    "ServerlessDeploymentBucket": {
      "Type": "AWS::S3::Bucket"
    },
    "HelloLogGroup": {
      "Type": "AWS::Logs::LogGroup",
      "Properties": {
        "LogGroupName": "/aws/lambda/timeseries-service-dev-hello"
      }
    },
    "IamRoleLambdaExecution": {
      "Type": "AWS::IAM::Role",
      "Properties": {
        "AssumeRolePolicyDocument": {
          "Version": "2012-10-17",
          "Statement": [
            {
              "Effect": "Allow",
              "Principal": {
                "Service": [
                  "lambda.amazonaws.com"
                ]
              },
              "Action": [
                "sts:AssumeRole"
              ]
            }
          ]
        },
        "Policies": [
          {
            "PolicyName": {
              "Fn::Join": [
                "-",
                [
                  "dev",
                  "timeseries-service",
                  "lambda"
                ]
              ]
            },
            "PolicyDocument": {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Action": [
                    "logs:CreateLogStream"
                  ],
                  "Resource": [
                    {
                      "Fn::Sub": "arn:${AWS::Partition}:logs:${AWS::Region}:${AWS::AccountId}:log-group:/aws/lambda/timeseries-service-dev-hello:*"
                    }
                  ]
                },
                {
                  "Effect": "Allow",
                  "Action": [
                    "logs:PutLogEvents"
                  ],
                  "Resource": [
                    {
                      "Fn::Sub": "arn:${AWS::Partition}:logs:${AWS::Region}:${AWS::AccountId}:log-group:/aws/lambda/timeseries-service-dev-hello:*:*"
                    }
                  ]
                }
              ]
            }
          }
        ],
        "Path": "/",
        "RoleName": {
          "Fn::Join": [
            "-",
            [
              "timeseries-service",
              "dev",
              "us-east-1",
              "lambdaRole"
            ]
          ]
        }
      }
    },
    "HelloLambdaFunction": {
      "Type": "AWS::Lambda::Function",
      "Properties": {
        "Code": {
          "S3Bucket": {
            "Ref": "ServerlessDeploymentBucket"
          },
          "S3Key": "serverless/timeseries-service/dev/1542744572309-2018-11-20T20:09:32.309Z/timeseries-service.zip"
        },
        "FunctionName": "timeseries-service-dev-hello",
        "Handler": "handler.hello",
        "MemorySize": 1024,
        "Role": {
          "Fn::GetAtt": [
            "IamRoleLambdaExecution",
            "Arn"
          ]
        },
        "Runtime": "python3.11",
        "Timeout": 6
      },
      "DependsOn": [
        "HelloLogGroup",
        "IamRoleLambdaExecution"
      ]
    },
    "HelloLambdaVersionU88Ag36tX5K6Yuze3R8jedH2g7q2TTGuafWQxEnUmo": {
      "Type": "AWS::Lambda::Version",
      "DeletionPolicy": "Retain",
      "Properties": {
        "FunctionName": {
          "Ref": "HelloLambdaFunction"
        },
        "CodeSha256": "+pq+8RveA979z1DNF8UKnFGZfgE07blNyJGust5VJnU="
      }
    },
    "ApiGatewayRestApi": {
      "Type": "AWS::ApiGateway::RestApi",
      "Properties": {
        "Name": "dev-timeseries-service",
        "EndpointConfiguration": {
          "Types": [
            "EDGE"
          ]
        }
      }
    },
    "ApiGatewayResourceHello": {
      "Type": "AWS::ApiGateway::Resource",
      "Properties": {
        "ParentId": {
          "Fn::GetAtt": [
            "ApiGatewayRestApi",
            "RootResourceId"
          ]
        },
        "PathPart": "hello",
        "RestApiId": {
          "Ref": "ApiGatewayRestApi"
        }
      }
    },
    "ApiGatewayMethodHelloGet": {
      "Type": "AWS::ApiGateway::Method",
      "Properties": {
        "HttpMethod": "GET",
        "RequestParameters": {},
        "ResourceId": {
          "Ref": "ApiGatewayResourceHello"
        },
        "RestApiId": {
          "Ref": "ApiGatewayRestApi"
        },
        "ApiKeyRequired": false,
        "AuthorizationType": "NONE",
        "Integration": {
          "IntegrationHttpMethod": "POST",
          "Type": "AWS_PROXY",
          "Uri": {
            "Fn::Join": [
              "",
              [
                "arn:",
                {
                  "Ref": "AWS::Partition"
                },
                ":apigateway:",
                {
                  "Ref": "AWS::Region"
                },
                ":lambda:path/2015-03-31/functions/",
                {
                  "Fn::GetAtt": [
                    "HelloLambdaFunction",
                    "Arn"
                  ]
                },
                "/invocations"
              ]
            ]
          }
        },
        "MethodResponses": []
      }
    },
    "ApiGatewayDeployment1542744572805": {
      "Type": "AWS::ApiGateway::Deployment",
      "Properties": {
        "RestApiId": {
          "Ref": "ApiGatewayRestApi"
        },
        "StageName": "dev"
      },
      "DependsOn": [
        "ApiGatewayMethodHelloGet"
      ]
    },
    "HelloLambdaPermissionApiGateway": {
      "Type": "AWS::Lambda::Permission",
      "Properties": {
        "FunctionName": {
          "Fn::GetAtt": [
            "HelloLambdaFunction",
            "Arn"
          ]
        },
        "Action": "lambda:InvokeFunction",
        "Principal": {
          "Fn::Join": [
            "",
            [
              "apigateway.",
              {
                "Ref": "AWS::URLSuffix"
              }
            ]
          ]
        },
        "SourceArn": {
          "Fn::Join": [
            "",
            [
              "arn:",
              {
                "Ref": "AWS::Partition"
              },
              ":execute-api:",
              {
                "Ref": "AWS::Region"
              },
              ":",
              {
                "Ref": "AWS::AccountId"
              },
              ":",
              {
                "Ref": "ApiGatewayRestApi"
              },
              "/*/*"
            ]
          ]
        }
      }
    }
  },
  "Outputs": {
    "ServerlessDeploymentBucketName": {
      "Value": {
        "Ref": "ServerlessDeploymentBucket"
      }
    },
    "HelloLambdaFunctionQualifiedArn": {
      "Description": "Current Lambda function version",
      "Value": {
        "Ref": "HelloLambdaVersionU88Ag36tX5K6Yuze3R8jedH2g7q2TTGuafWQxEnUmo"
      }
    },
    "ServiceEndpoint": {
      "Description": "URL of the service endpoint",
      "Value": {
        "Fn::Join": [
          "",
          [
            "https://",
            {
              "Ref": "ApiGatewayRestApi"
            },
            ".execute-api.us-east-1.",
            {
              "Ref": "AWS::URLSuffix"
            },
            "/dev"
          ]
        ]
      }
    }
  }
}"""

template_with_missing_sub = """{
  "AWSTemplateFormatVersion": "2010-09-09",
  "Description": "AWS CloudFormation template",
  "Resources": {
    "UnknownResource": {"Type": "AWS::Cloud9::EnvironmentEC2", "Properties": {}},
    "ApiGatewayRestApi": {
      "Type": "AWS::ApiGateway::RestApi",
      "Properties": {
        "Name": "test-api",
        "Description": {"Fn::Sub": "${UnknownResource}"},
        "EndpointConfiguration": {
          "Types": [
            "EDGE"
          ]
        }
      }
    }
  },
  "Outputs": {
  }
}"""


@mock_aws
def test_simple_apigateway_with_lambda_proxy():
    region = "us-east-1"
    apigw = boto3.client("apigateway", region_name=region)
    cf = boto3.client("cloudformation", region_name=region)
    awslambda = boto3.client("lambda", region_name=region)
    cf.create_stack(StackName="teststack", TemplateBody=template)
    #
    cf.describe_stacks(StackName="teststack")["Stacks"]
    resources = cf.describe_stack_resources(StackName="teststack")["StackResources"]
    api_id = [
        r["PhysicalResourceId"]
        for r in resources
        if r["ResourceType"] == "AWS::ApiGateway::RestApi"
    ][0]
    fn_name = [
        r["PhysicalResourceId"]
        for r in resources
        if r["LogicalResourceId"] == "HelloLambdaFunction"
    ][0]
    #
    # Verify Rest API was created
    api = apigw.get_rest_apis()["items"][0]
    assert api["id"] == api_id
    assert api["name"] == "dev-timeseries-service"
    #
    # Verify Gateway Resource was created
    paths = apigw.get_resources(restApiId=api_id)["items"]
    root_path = [p for p in paths if p["path"] == "/"][0]
    hello_path = [p for p in paths if p["path"] == "/hello"][0]
    assert hello_path["parentId"] == root_path["id"]
    #
    # Verify Gateway Method was created
    m = apigw.get_method(
        restApiId=api_id, resourceId=hello_path["id"], httpMethod="GET"
    )
    assert m["httpMethod"] == "GET"
    #
    # Verify a Gateway Deployment was created
    d = apigw.get_deployments(restApiId=api_id)["items"]
    assert len(d) == 1
    #
    # Verify Lambda function was created
    awslambda.get_function(FunctionName=fn_name)  # Will throw 404 if it doesn't exist
    #
    # Verify Lambda Permission was created
    policy = json.loads(awslambda.get_policy(FunctionName=fn_name)["Policy"])
    statement = policy["Statement"][0]
    assert (
        statement["FunctionName"]
        == f"arn:aws:lambda:us-east-1:123456789012:function:{fn_name}"
    )
    assert (
        statement["Condition"]["ArnLike"]["AWS:SourceArn"]
        == f"arn:aws:execute-api:us-east-1:123456789012:{api_id}/*/*"
    )


@mock_aws
@pytest.mark.filterwarnings("ignore:Tried to parse")
def test_apigateway_with_unknown_description():
    region = "us-east-1"
    apigw = boto3.client("apigateway", region_name=region)
    cf = boto3.client("cloudformation", region_name=region)
    cf.create_stack(StackName="teststack", TemplateBody=template_with_missing_sub)

    api = apigw.get_rest_apis()["items"][0]
    assert api["description"] == "${UnknownResource}"



"""Unit tests for osis-supported APIs."""

import json

import boto3
import pytest
import requests
from botocore.exceptions import ClientError

from moto import mock_aws, settings
from moto.moto_api import state_manager

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html

BASIC_PIPELINE_KWARGS = {
    "PipelineName": "test",
    "MinUnits": 2,
    "MaxUnits": 4,
    "PipelineConfigurationBody": """version: "2"\nopensearch-migration-pipeline:\n  source:
        \n    opensearch:\n      acknowledgments: true
        \n      hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n      indices:\n        exclude:\n          - index_name_regex: \'\\..*\'\n      aws:
        \n        region: "eu-west-1"\n        sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n        serverless: false\n  sink:\n    - opensearch:
        \n        hosts: ["https://kbjahvxo2jgx8beq2vob.eu-west-1.aoss.amazonaws.com"]
        \n        aws:\n          sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n          region: "eu-west-1"\n          serverless: true\n""",
}


@mock_aws
def test_create_pipeline():
    set_transition()
    client = boto3.client("osis", region_name="eu-west-1")
    resp = client.create_pipeline(**BASIC_PIPELINE_KWARGS)["Pipeline"]
    assert resp["PipelineName"] == "test"
    assert resp["PipelineArn"] == "arn:aws:osis:eu-west-1:123456789012:pipeline/test"
    assert resp["MinUnits"] == 2
    assert resp["MaxUnits"] == 4
    assert resp["Status"] == "ACTIVE"
    assert (
        resp["StatusReason"]["Description"] == "The pipeline is ready to ingest data."
    )
    assert (
        resp["PipelineConfigurationBody"]
        == BASIC_PIPELINE_KWARGS["PipelineConfigurationBody"]
    )
    assert (
        ".eu-west-1.osis.amazonaws.com" in resp["IngestEndpointUrls"][0]
        and "test" in resp["IngestEndpointUrls"][0]
    )
    assert resp["ServiceVpcEndpoints"][0]["ServiceName"] == "OPENSEARCH_SERVERLESS"
    assert resp["Destinations"][0]["ServiceName"] == "OpenSearch_Serverless"
    assert (
        resp["Destinations"][0]["Endpoint"]
        == "https://kbjahvxo2jgx8beq2vob.eu-west-1.aoss.amazonaws.com"
    )
    assert "VpcEndpointService" not in resp
    assert "VpcOptions" not in resp
    assert resp["Tags"] == []

    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="172.28.7.192/26")
    sg = ec2.create_security_group(
        GroupName="test-group", Description="Test security group sg01"
    )

    kwargs = {
        "PipelineName": "test-2",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": """log-pipeline:\n  processor:\n  - date:
        \n      destination: \'@timestamp\'\n      from_time_received: true\n  - delete_entries:
        \n      with_keys:\n      - s3\n  sink:\n  - opensearch:\n      aws:
        \n        region: eu-west-1\n        serverless: true
        \n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole\n      hosts:
        \n      - https://kbjahvxo2jgx8beq2vob.eu-west-1.aoss.amazonaws.com
        \n      index: uncompressed_logs\n  - opensearch:\n      aws:\n        region: eu-west-1
        \n        serverless: false\n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole
        \n      hosts:
        \n      - https://vpc-c7ntest-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com
        \n      index: uncompressed_logs\n      dlq:\n        s3:\n          bucket: "dlq-bucket"
        \n          region: "eu-west-1"
        \n          sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"\n  - s3:\n      aws:
        \n        region: eu-west-1\n        serverless: false
        \n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole
        \n      bucket: test-s3-bucket-2\n      threshold: 3\n      codec: json\n  - s3:
        \n      aws:\n        region: eu-west-1\n        serverless: false
        \n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole
        \n      bucket: test-s3-bucket-1\n      threshold: 3\n      codec: json\n  source:
        \n    s3:\n      acknowledgments: true\n      aws:\n        region: eu-west-1
        \n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole\n      codec:
        \n        newline: null\n      compression: none\n      scan:\n        buckets:
        \n        - bucket:\n            name: test-s3-bucket-2\nlog-pipeline-2:\n  processor:
        \n  - date:\n      destination: \'@timestamp\'\n      from_time_received: true
        \n  - delete_entries:\n      with_keys:\n      - s3\n  sink:\n  - pipeline:
        \n      name: "log-to-metrics-pipeline"\n  - opensearch:\n      aws:
        \n        region: eu-west-1\n        serverless: false
        \n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole\n      hosts:
        \n      - https://vpc-c7ntest-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com
        \n      index: uncompressed_logs\n  - s3:\n      aws:\n        region: eu-west-1
        \n        serverless: false\n        sts_role_arn: arn:aws:iam::123456789012:role/MyRole
        \n      bucket: test-s3-bucket-1\n      threshold: 3\n      codec: json\n  source:
        \n    pipeline:\n      name: "apache-log-pipeline-with-metrics"\nversion: \'2\'""",
        "LogPublishingOptions": {
            "IsLoggingEnabled": True,
            "CloudWatchLogDestination": {
                "LogGroup": "/aws/osis/test",
            },
        },
        "VpcOptions": {
            "SubnetIds": [subnet.id],
            "SecurityGroupIds": [sg.id],
            "VpcEndpointManagement": "SERVICE",
            "VpcAttachmentOptions": {
                "AttachToVpc": True,
                "CidrBlock": "172.168.1.1",
            },
        },
        "BufferOptions": {
            "PersistentBufferEnabled": True,
        },
        "EncryptionAtRestOptions": {
            "KmsKeyArn": "arn:aws:kms:eu-west-1:123456789012:key/12345678-1234-1234-1234-123456789012",
        },
        "Tags": [
            {
                "Key": "TestKey",
                "Value": "TestValue",
            }
        ],
    }
    resp = client.create_pipeline(**kwargs)["Pipeline"]
    assert resp["PipelineName"] == "test-2"
    assert "CreatedAt" in resp
    assert "LastUpdatedAt" in resp
    assert resp["LogPublishingOptions"]["IsLoggingEnabled"]
    assert (
        resp["LogPublishingOptions"]["CloudWatchLogDestination"]["LogGroup"]
        == "/aws/osis/test"
    )
    assert resp["VpcEndpoints"][0]["VpcOptions"]["SubnetIds"] == [subnet.id]
    assert resp["VpcEndpoints"][0]["VpcOptions"]["SecurityGroupIds"] == [sg.id]
    assert resp["VpcEndpoints"][0]["VpcOptions"]["VpcEndpointManagement"] == "SERVICE"
    assert resp["VpcEndpoints"][0]["VpcOptions"]["VpcAttachmentOptions"]["AttachToVpc"]
    assert (
        resp["VpcEndpoints"][0]["VpcOptions"]["VpcAttachmentOptions"]["CidrBlock"]
        == "172.168.1.1"
    )
    assert (
        resp["VpcEndpoints"][0]["VpcEndpointId"]
        and resp["VpcEndpoints"][0]["VpcId"] == vpc.id
    )
    assert resp["BufferOptions"]["PersistentBufferEnabled"]
    assert (
        resp["EncryptionAtRestOptions"]["KmsKeyArn"]
        == "arn:aws:kms:eu-west-1:123456789012:key/12345678-1234-1234-1234-123456789012"
    )
    assert "VpcEndpointService" not in resp
    assert resp["ServiceVpcEndpoints"][0]["ServiceName"] == "OPENSEARCH_SERVERLESS"
    assert resp["Destinations"] == [
        {
            "ServiceName": "OpenSearch_Serverless",
            "Endpoint": "https://kbjahvxo2jgx8beq2vob.eu-west-1.aoss.amazonaws.com",
        },
        {
            "ServiceName": "OpenSearch",
            "Endpoint": "https://vpc-c7ntest-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com",
        },
        {"ServiceName": "S3", "Endpoint": "test-s3-bucket-2"},
        {"ServiceName": "S3", "Endpoint": "test-s3-bucket-1"},
        {
            "ServiceName": "OpenSearch",
            "Endpoint": "https://vpc-c7ntest-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com",
        },
        {"ServiceName": "S3", "Endpoint": "test-s3-bucket-1"},
    ]

    assert (
        resp["Tags"][0]["Key"] == "TestKey" and resp["Tags"][0]["Value"] == "TestValue"
    )

    assert (
        boto3.client("ec2", region_name="eu-west-1").describe_vpc_endpoints(
            VpcEndpointIds=[resp["VpcEndpoints"][0]["VpcEndpointId"]]
        )["VpcEndpoints"]
        != []
    )


@mock_aws
def test_create_pipeline_customer_endpoint():
    set_transition({"progression": "manual", "times": 1})
    client = boto3.client("osis", region_name="eu-west-1")
    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="172.28.7.192/26")
    sg = ec2.create_security_group(
        GroupName="test-group", Description="Test security group sg01"
    )
    kwargs = {
        "PipelineName": "test",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": """version: "2"\nopensearch-migration-pipeline:\n  source:
        \n    opensearch:\n      acknowledgments: true
        \n      hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n      indices:\n        exclude:\n          - index_name_regex: \'\\..*\'\n      aws:
        \n        region: "eu-west-1"\n        sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n        serverless: false\n  sink:\n    - opensearch:
        \n        hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n        aws:\n          sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n          region: "eu-west-1"\n          serverless: false\n""",
        "VpcOptions": {
            "SubnetIds": [subnet.id],
            "SecurityGroupIds": [sg.id],
            "VpcEndpointManagement": "CUSTOMER",
        },
    }
    resp = client.create_pipeline(**kwargs)["Pipeline"]
    assert resp["PipelineName"] == "test"
    assert "VpcEndpointService" in resp
    assert "ServiceVpcEndpoints" not in resp
    assert resp["VpcEndpoints"][0]["VpcOptions"]["VpcEndpointManagement"] == "CUSTOMER"
    assert "VpcEndpointId" not in resp["VpcEndpoints"][0]
    assert "VpcAttachmentOptions" not in resp["VpcEndpoints"][0]
    assert resp["Status"] == "CREATING"


@mock_aws
def test_create_pipeline_error():
    set_transition()
    client = boto3.client("osis", region_name="eu-west-1")
    kwargs = {
        "PipelineName": "test",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": BASIC_PIPELINE_KWARGS["PipelineConfigurationBody"],
    }
    client.create_pipeline(**kwargs)["Pipeline"]
    with pytest.raises(ClientError) as exc:
        client.create_pipeline(**kwargs)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceAlreadyExistsException"

    kwargs["PipelineName"] = "test-2"
    kwargs["VpcOptions"] = {}
    kwargs["VpcOptions"]["SubnetIds"] = ["subnet-12345678901234567"]

    with pytest.raises(ClientError) as exc:
        client.create_pipeline(**kwargs)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "Invalid VpcOptions: The subnet ID subnet-12345678901234567 does not exist"
    )

    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="172.28.7.192/26")

    kwargs["VpcOptions"]["SubnetIds"] = [subnet.id]
    kwargs["VpcOptions"]["SecurityGroupIds"] = ["sg-12345678901234567"]

    with pytest.raises(ClientError) as exc:
        client.create_pipeline(**kwargs)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "Invalid VpcOptions: The security group sg-12345678901234567 does not exist"
    )

    kwargs["VpcOptions"].pop("SecurityGroupIds")
    vpc = ec2.create_vpc(CidrBlock="172.29.7.0/24")
    subnet_2 = ec2.create_subnet(VpcId=vpc.id, CidrBlock="172.29.7.192/26")
    kwargs["VpcOptions"]["SubnetIds"].append(subnet_2.id)

    with pytest.raises(ClientError) as exc:
        client.create_pipeline(**kwargs)
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"
    assert (
        err["Message"]
        == "Invalid VpcOptions: All specified subnets must belong to the same VPC."
    )


@mock_aws
def test_update_pipeline():
    set_transition()
    client = boto3.client("osis", region_name="eu-west-1")
    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="172.28.7.192/26")
    sg = ec2.create_security_group(
        GroupName="test-group", Description="Test security group sg01"
    )

    kwargs = {
        "PipelineName": "test",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": BASIC_PIPELINE_KWARGS["PipelineConfigurationBody"],
        "LogPublishingOptions": {
            "IsLoggingEnabled": True,
            "CloudWatchLogDestination": {
                "LogGroup": "/aws/osis/test",
            },
        },
        "VpcOptions": {
            "SubnetIds": [subnet.id],
            "SecurityGroupIds": [sg.id],
            "VpcEndpointManagement": "SERVICE",
            "VpcAttachmentOptions": {
                "AttachToVpc": True,
                "CidrBlock": "172.168.1.1",
            },
        },
        "BufferOptions": {
            "PersistentBufferEnabled": True,
        },
        "EncryptionAtRestOptions": {
            "KmsKeyArn": "arn:aws:kms:eu-west-1:123456789012:key/12345678-1234-1234-1234-123456789012",
        },
        "Tags": [
            {
                "Key": "TestKey",
                "Value": "TestValue",
            }
        ],
    }
    original = client.create_pipeline(**kwargs)["Pipeline"]
    resp = client.update_pipeline(PipelineName="test", MinUnits=3)["Pipeline"]
    assert resp["MinUnits"] == 3
    assert resp["MaxUnits"] == original["MaxUnits"]
    assert (
        resp["PipelineConfigurationBody"]
        == BASIC_PIPELINE_KWARGS["PipelineConfigurationBody"]
    )
    assert resp["Destinations"] == original["Destinations"]
    assert resp["ServiceVpcEndpoints"] == original["ServiceVpcEndpoints"]
    assert resp["VpcEndpoints"] == original["VpcEndpoints"]
    assert resp["Tags"] == original["Tags"]
    assert resp["LogPublishingOptions"] == original["LogPublishingOptions"]
    assert resp["BufferOptions"] == original["BufferOptions"]
    assert resp["EncryptionAtRestOptions"] == original["EncryptionAtRestOptions"]
    assert resp["Status"] == "ACTIVE"


@mock_aws
def test_update_pipeline_all_args():
    set_transition({"progression": "manual", "times": 1})
    client = boto3.client("osis", region_name="eu-west-1")
    kwargs = {
        "PipelineName": "test",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": BASIC_PIPELINE_KWARGS["PipelineConfigurationBody"],
        "LogPublishingOptions": {
            "IsLoggingEnabled": True,
            "CloudWatchLogDestination": {
                "LogGroup": "test/osis/logs",
            },
        },
        "BufferOptions": {
            "PersistentBufferEnabled": False,
        },
        "EncryptionAtRestOptions": {
            "KmsKeyArn": "arn:aws:kms:eu-west-1:123456789012:key/12345678-1234-1234-1234-123456789012",
        },
    }

    resp = client.create_pipeline(**kwargs)["Pipeline"]
    last_updated = resp["LastUpdatedAt"]
    # state transition
    client.list_pipelines()

    new_pipeline_config = """version: "2"\nopensearch-migration-pipeline:\n  source:
        \n    opensearch:\n      acknowledgments: true
        \n      hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n      indices:\n        exclude:\n          - index_name_regex: \'\\..*\'\n      aws:
        \n        region: "eu-west-1"\n        sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n        serverless: false\n  sink:\n    - opensearch:
        \n        hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n        aws:\n          sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n          region: "eu-west-1"\n          serverless: false\n"""
    new_key = (
        "arn:aws:kms:eu-west-1:123456789012:key/87654321-4321-4321-4321-210987654321"
    )
    kwargs["MinUnits"] = 3
    kwargs["MaxUnits"] = 5
    kwargs["PipelineConfigurationBody"] = new_pipeline_config
    kwargs["LogPublishingOptions"]["IsLoggingEnabled"] = False
    kwargs["BufferOptions"]["PersistentBufferEnabled"] = True
    kwargs["EncryptionAtRestOptions"]["KmsKeyArn"] = new_key

    resp = client.update_pipeline(**kwargs)["Pipeline"]

    assert resp["MinUnits"] == 3
    assert resp["MaxUnits"] == 5
    assert resp["PipelineConfigurationBody"] == new_pipeline_config
    assert not resp["LogPublishingOptions"]["IsLoggingEnabled"]
    assert resp["BufferOptions"]["PersistentBufferEnabled"]
    assert resp["EncryptionAtRestOptions"]["KmsKeyArn"] == new_key
    assert resp["Destinations"][0]["ServiceName"] == "OpenSearch"
    assert (
        resp["Destinations"][0]["Endpoint"]
        == "https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"
    )
    assert "ServiceVpcEndpoints" not in resp
    assert resp["LastUpdatedAt"] > last_updated
    assert resp["Status"] == "UPDATING"


@mock_aws
def test_update_pipeline_error():
    set_transition({"progression": "manual", "times": 1})
    client = boto3.client("osis", region_name="eu-west-1")
    client.create_pipeline(**BASIC_PIPELINE_KWARGS)["Pipeline"]
    with pytest.raises(ClientError) as exc:
        client.update_pipeline(PipelineName="test", MinUnits=3)["Pipeline"]
    err = exc.value.response["Error"]
    assert err["Code"] == "ConflictException"
    assert (
        err["Message"]
        == "Only pipelines with one of the following statuses are eligible for updates: ['UPDATE_FAILED', 'ACTIVE', 'START_FAILED', 'STOPPED']. The current status is CREATING."
    )

    with pytest.raises(ClientError) as exc:
        client.update_pipeline(PipelineName="test-2", MinUnits=3)["Pipeline"]
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == "Pipeline test-2 could not be found."


@mock_aws
def test_delete_pipeline():
    set_transition({"progression": "manual", "times": 2})
    client = boto3.client("osis", region_name="eu-west-1")
    original = client.create_pipeline(**BASIC_PIPELINE_KWARGS)["Pipeline"]
    for _ in range(2):
        client.list_pipelines()
    client.delete_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    pipeline = client.list_pipelines()["Pipelines"][0]
    assert pipeline["PipelineName"] == BASIC_PIPELINE_KWARGS["PipelineName"]
    assert pipeline["Status"] == "DELETING"
    assert pipeline["LastUpdatedAt"] > original["LastUpdatedAt"]
    pipelines = client.list_pipelines()["Pipelines"]
    assert pipelines == []


@mock_aws
def test_delete_pipeline_error():
    set_transition({"progression": "manual", "times": 1})
    client = boto3.client("osis", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.delete_pipeline(PipelineName="test")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == "Pipeline test could not be found."

    client.create_pipeline(**BASIC_PIPELINE_KWARGS)["Pipeline"]
    with pytest.raises(ClientError) as exc:
        client.delete_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    err = exc.value.response["Error"]
    assert err["Code"] == "ConflictException"
    assert (
        err["Message"]
        == "Only pipelines with one of the following statuses are eligible for deletion: ['UPDATE_FAILED', 'ACTIVE', 'START_FAILED', 'STOPPED', 'CREATE_FAILED']. The current status is CREATING."
    )


@mock_aws
def test_get_pipeline():
    client = boto3.client("osis", region_name="eu-west-1")
    client.create_pipeline(**BASIC_PIPELINE_KWARGS)["Pipeline"]
    resp = client.get_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])[
        "Pipeline"
    ]
    assert resp["PipelineName"] == BASIC_PIPELINE_KWARGS["PipelineName"]
    assert (
        resp["PipelineArn"]
        == f"arn:aws:osis:eu-west-1:123456789012:pipeline/{BASIC_PIPELINE_KWARGS['PipelineName']}"
    )
    assert resp["MinUnits"] == BASIC_PIPELINE_KWARGS["MinUnits"]
    assert resp["MaxUnits"] == BASIC_PIPELINE_KWARGS["MaxUnits"]
    assert resp["Status"] == "ACTIVE"
    assert (
        resp["StatusReason"]["Description"] == "The pipeline is ready to ingest data."
    )
    assert (
        resp["PipelineConfigurationBody"]
        == BASIC_PIPELINE_KWARGS["PipelineConfigurationBody"]
    )
    assert (
        ".eu-west-1.osis.amazonaws.com" in resp["IngestEndpointUrls"][0]
        and BASIC_PIPELINE_KWARGS["PipelineName"] in resp["IngestEndpointUrls"][0]
    )
    assert resp["ServiceVpcEndpoints"][0]["ServiceName"] == "OPENSEARCH_SERVERLESS"
    assert resp["Destinations"][0]["ServiceName"] == "OpenSearch_Serverless"
    assert (
        resp["Destinations"][0]["Endpoint"]
        == "https://kbjahvxo2jgx8beq2vob.eu-west-1.aoss.amazonaws.com"
    )
    assert "VpcEndpointService" not in resp
    assert "VpcEndpoints" not in resp
    assert resp["Tags"] == []
    assert "CreatedAt" in resp
    assert "LastUpdatedAt" in resp
    assert "LogPublishingOptions" not in resp
    assert "BufferOptions" not in resp
    assert "EncryptionAtRestOptions" not in resp

    ec2 = boto3.resource("ec2", region_name="eu-west-1")
    vpc = ec2.create_vpc(CidrBlock="172.28.7.0/24")
    subnet = ec2.create_subnet(VpcId=vpc.id, CidrBlock="172.28.7.192/26")
    sg = ec2.create_security_group(
        GroupName="test-group", Description="Test security group sg01"
    )

    kwargs = {
        "PipelineName": "test-2",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": """version: "2"\nopensearch-migration-pipeline:\n  source:
        \n    opensearch:\n      acknowledgments: true
        \n      hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n      indices:\n        exclude:\n          - index_name_regex: \'\\..*\'\n      aws:
        \n        region: "eu-west-1"\n        sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n        serverless: false\n  sink:\n    - opensearch:
        \n        hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
        \n        aws:\n          sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
        \n          region: "eu-west-1"\n          serverless: false\n""",
        "LogPublishingOptions": {
            "IsLoggingEnabled": True,
            "CloudWatchLogDestination": {
                "LogGroup": "/aws/osis/test",
            },
        },
        "VpcOptions": {
            "SubnetIds": [subnet.id],
            "SecurityGroupIds": [sg.id],
            "VpcEndpointManagement": "SERVICE",
            "VpcAttachmentOptions": {
                "AttachToVpc": True,
                "CidrBlock": "172.168.1.1",
            },
        },
        "BufferOptions": {
            "PersistentBufferEnabled": True,
        },
        "EncryptionAtRestOptions": {
            "KmsKeyArn": "arn:aws:kms:eu-west-1:123456789012:key/12345678-1234-1234-1234-123456789012",
        },
        "Tags": [
            {
                "Key": "TestKey",
                "Value": "TestValue",
            }
        ],
    }
    client.create_pipeline(**kwargs)
    resp = client.get_pipeline(PipelineName="test-2")["Pipeline"]
    assert "ServiceVpcEndpoints" not in resp
    assert resp["Destinations"][0]["ServiceName"] == "OpenSearch"
    assert (
        resp["Destinations"][0]["Endpoint"]
        == "https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"
    )
    assert "VpcEndpointService" not in resp
    assert resp["VpcEndpoints"][0]["VpcOptions"] == {
        "SubnetIds": [subnet.id],
        "SecurityGroupIds": [sg.id],
        "VpcEndpointManagement": "SERVICE",
        "VpcAttachmentOptions": {
            "AttachToVpc": True,
            "CidrBlock": "172.168.1.1",
        },
    }
    assert "VpcEndpointId" in resp["VpcEndpoints"][0]
    assert resp["VpcEndpoints"][0]["VpcId"] == vpc.id
    assert resp["Tags"] == [
        {
            "Key": "TestKey",
            "Value": "TestValue",
        }
    ]
    assert resp["LogPublishingOptions"] == {
        "IsLoggingEnabled": True,
        "CloudWatchLogDestination": {
            "LogGroup": "/aws/osis/test",
        },
    }
    assert resp["BufferOptions"]["PersistentBufferEnabled"]
    assert (
        resp["EncryptionAtRestOptions"]["KmsKeyArn"]
        == "arn:aws:kms:eu-west-1:123456789012:key/12345678-1234-1234-1234-123456789012"
    )


@mock_aws
def test_get_pipeline_error():
    client = boto3.client("osis", region_name="eu-west-1")
    with pytest.raises(ClientError) as exc:
        client.get_pipeline(PipelineName="test")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == "Pipeline test could not be found."


@mock_aws
def test_list_pipelines():
    set_transition({"progression": "manual", "times": 1})
    client = boto3.client("osis", region_name="ap-southeast-1")
    resp = client.list_pipelines()
    assert resp["Pipelines"] == []
    client.create_pipeline(**BASIC_PIPELINE_KWARGS)
    resp = client.list_pipelines()["Pipelines"][0]
    assert resp["PipelineName"] == "test"
    assert (
        resp["PipelineArn"]
        == f"arn:aws:osis:ap-southeast-1:123456789012:pipeline/{BASIC_PIPELINE_KWARGS['PipelineName']}"
    )
    assert resp["MinUnits"] == 2
    assert resp["MaxUnits"] == 4
    assert resp["Status"] == "ACTIVE"
    assert (
        resp["StatusReason"]["Description"] == "The pipeline is ready to ingest data."
    )
    assert resp["Tags"] == []
    assert "CreatedAt" in resp
    assert "LastUpdatedAt" in resp

    kwargs = {
        "PipelineName": "test-2",
        "MinUnits": 2,
        "MaxUnits": 4,
        "PipelineConfigurationBody": """version: "2"\nopensearch-migration-pipeline:\n  source:
            \n    opensearch:\n      acknowledgments: true
            \n      hosts: ["https://vpc-test-ieeljhbsnht35i5rtzjl756pk4.eu-west-1.es.amazonaws.com"]
            \n      indices:\n        exclude:\n          - index_name_regex: \'\\..*\'\n      aws:
            \n        region: "eu-west-1"\n        sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
            \n        serverless: false\n  sink:\n    - opensearch:
            \n        hosts: ["https://kbjahvxo2jgx8beq2vob.eu-west-1.aoss.amazonaws.com"]
            \n        aws:\n          sts_role_arn: "arn:aws:iam::123456789012:role/MyRole"
            \n          region: "eu-west-1"\n          serverless: true\n""",
    }
    client.create_pipeline(**kwargs)
    assert len(client.list_pipelines()["Pipelines"]) == 2


@mock_aws
def test_list_tags_for_resource():
    client = boto3.client("osis", region_name="eu-west-1")
    resp = client.create_pipeline(
        **BASIC_PIPELINE_KWARGS, Tags=[{"Key": "TestKey", "Value": "TestValue"}]
    )["Pipeline"]
    tags = client.list_tags_for_resource(Arn=resp["PipelineArn"])["Tags"]
    assert tags[0]["Key"] == "TestKey"
    assert tags[0]["Value"] == "TestValue"


@mock_aws
def test_stop_pipeline():
    set_transition({"progression": "manual", "times": 2})
    client = boto3.client("osis", region_name="eu-west-1")
    client.create_pipeline(**BASIC_PIPELINE_KWARGS)

    with pytest.raises(ClientError) as exc:
        client.stop_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    err = exc.value.response["Error"]
    assert err["Code"] == "ConflictException"
    assert (
        err["Message"]
        == "Only pipelines with one of the following statuses are eligible for stopping: ['UPDATE_FAILED', 'ACTIVE']. The current status is CREATING."
    )

    for _ in range(2):
        client.list_pipelines()
    client.stop_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    pipeline = client.list_pipelines()["Pipelines"][0]
    assert pipeline["Status"] == "STOPPING"

    for _ in range(2):
        client.get_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    pipeline = client.list_pipelines()["Pipelines"][0]
    assert pipeline["Status"] == "STOPPED"


@mock_aws
def test_start_pipeline():
    set_transition({"progression": "manual", "times": 2})
    client = boto3.client("osis", region_name="eu-west-1")
    client.create_pipeline(**BASIC_PIPELINE_KWARGS)

    with pytest.raises(ClientError) as exc:
        client.start_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    err = exc.value.response["Error"]
    assert err["Code"] == "ConflictException"
    assert (
        err["Message"]
        == "Only pipelines with one of the following statuses are eligible for starting: ['START_FAILED', 'STOPPED']. The current status is CREATING."
    )

    for _ in range(2):
        client.list_pipelines()

    client.stop_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])

    for _ in range(2):
        client.list_pipelines()

    client.start_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])
    pipeline = client.list_pipelines()["Pipelines"][0]
    assert pipeline["Status"] == "STARTING"
    client.list_pipelines()
    pipeline = client.list_pipelines()["Pipelines"][0]
    assert pipeline["Status"] == "ACTIVE"


@mock_aws
def test_tag_resource():
    client = boto3.client("osis", region_name="eu-west-1")
    resp = client.create_pipeline(**BASIC_PIPELINE_KWARGS)["Pipeline"]
    client.tag_resource(
        Arn=resp["PipelineArn"], Tags=[{"Key": "TestKey", "Value": "TestValue"}]
    )
    resp = client.get_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])[
        "Pipeline"
    ]
    assert resp["Tags"] == [{"Key": "TestKey", "Value": "TestValue"}]


@mock_aws
def test_untag_resource():
    client = boto3.client("osis", region_name="eu-west-1")
    resp = client.create_pipeline(
        **BASIC_PIPELINE_KWARGS, Tags=[{"Key": "TestKey", "Value": "TestValue"}]
    )["Pipeline"]
    client.untag_resource(Arn=resp["PipelineArn"], TagKeys=["TestKey"])
    resp = client.get_pipeline(PipelineName=BASIC_PIPELINE_KWARGS["PipelineName"])[
        "Pipeline"
    ]
    assert resp["Tags"] == []


def set_transition(transition={"progression": "immediate"}):
    if settings.TEST_DECORATOR_MODE:
        state_manager.set_transition(model_name="osis::pipeline", transition=transition)
    else:
        post_body = dict(model_name="osis::pipeline", transition=transition)
        requests.post(
            "http://localhost:5000/moto-api/state-manager/set-transition",
            data=json.dumps(post_body),
        )



"""Handles incoming lexv2models requests, invokes methods, returns responses."""

import json
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import LexModelsV2Backend, lexv2models_backends


class LexModelsV2Response(BaseResponse):
    """Handler for LexModelsV2 requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="lexv2-models")

    @property
    def lexv2models_backend(self) -> LexModelsV2Backend:
        """Return backend instance specific for this region."""
        # lexv2models_backends is not yet typed
        # Please modify moto/backends.py to add the appropriate type annotations for this service
        return lexv2models_backends[self.current_account][self.region]

    def create_bot(self) -> str:
        bot_name = self._get_param("botName")
        description = self._get_param("description")
        role_arn = self._get_param("roleArn")
        data_privacy = self._get_param("dataPrivacy")
        idle_session_ttl_in_seconds = self._get_param("idleSessionTTLInSeconds")
        bot_tags = self._get_param("botTags")
        test_bot_alias_tags = self._get_param("testBotAliasTags")
        bot_type = self._get_param("botType")
        bot_members = self._get_param("botMembers")

        resp = self.lexv2models_backend.create_bot(
            bot_name=bot_name,
            description=description,
            role_arn=role_arn,
            data_privacy=data_privacy,
            idle_session_ttl_in_seconds=idle_session_ttl_in_seconds,
            bot_tags=bot_tags,
            test_bot_alias_tags=test_bot_alias_tags,
            bot_type=bot_type,
            bot_members=bot_members,
        )

        return json.dumps(resp)

    def describe_bot(self) -> str:
        bot_id = self._get_param("botId")

        return json.dumps(
            self.lexv2models_backend.describe_bot(
                bot_id=bot_id,
            )
        )

    def update_bot(self) -> str:
        bot_id = self._get_param("botId")
        bot_name = self._get_param("botName")
        description = self._get_param("description")
        role_arn = self._get_param("roleArn")
        data_privacy = self._get_param("dataPrivacy")
        idle_session_ttl_in_seconds = self._get_param("idleSessionTTLInSeconds")
        bot_type = self._get_param("botType")
        bot_members = self._get_param("botMembers")

        return json.dumps(
            self.lexv2models_backend.update_bot(
                bot_id=bot_id,
                bot_name=bot_name,
                description=description,
                role_arn=role_arn,
                data_privacy=data_privacy,
                idle_session_ttl_in_seconds=idle_session_ttl_in_seconds,
                bot_type=bot_type,
                bot_members=bot_members,
            )
        )

    def list_bots(self) -> str:
        bot_summaries = self.lexv2models_backend.list_bots()
        return json.dumps(dict(botSummaries=bot_summaries, nextToken=None))

    def delete_bot(self) -> str:
        bot_id = self._get_param("botId")
        skip_resource_in_use_check = self._get_param("skipResourceInUseCheck")
        bot_id, bot_status = self.lexv2models_backend.delete_bot(
            bot_id=bot_id,
            skip_resource_in_use_check=skip_resource_in_use_check,
        )
        return json.dumps(dict(botId=bot_id, botStatus=bot_status))

    def create_bot_alias(self) -> str:
        bot_alias_name = self._get_param("botAliasName")
        description = self._get_param("description")
        bot_version = self._get_param("botVersion")
        bot_alias_locale_settings = self._get_param("botAliasLocaleSettings")
        conversation_log_settings = self._get_param("conversationLogSettings")
        sentiment_analysis_settings = self._get_param("sentimentAnalysisSettings")
        bot_id = self._get_param("botId")
        tags = self._get_param("tags")

        return json.dumps(
            self.lexv2models_backend.create_bot_alias(
                bot_alias_name=bot_alias_name,
                description=description,
                bot_version=bot_version,
                bot_alias_locale_settings=bot_alias_locale_settings,
                conversation_log_settings=conversation_log_settings,
                sentiment_analysis_settings=sentiment_analysis_settings,
                bot_id=bot_id,
                tags=tags,
            )
        )

    def describe_bot_alias(self) -> str:
        bot_alias_id = self._get_param("botAliasId")
        bot_id = self._get_param("botId")

        return json.dumps(
            self.lexv2models_backend.describe_bot_alias(
                bot_alias_id=bot_alias_id,
                bot_id=bot_id,
            )
        )

    def update_bot_alias(self) -> str:
        bot_alias_id = self._get_param("botAliasId")
        bot_alias_name = self._get_param("botAliasName")
        description = self._get_param("description")
        bot_version = self._get_param("botVersion")
        bot_alias_locale_settings = self._get_param("botAliasLocaleSettings")
        conversation_log_settings = self._get_param("conversationLogSettings")
        sentiment_analysis_settings = self._get_param("sentimentAnalysisSettings")
        bot_id = self._get_param("botId")

        return json.dumps(
            self.lexv2models_backend.update_bot_alias(
                bot_alias_id=bot_alias_id,
                bot_alias_name=bot_alias_name,
                description=description,
                bot_version=bot_version,
                bot_alias_locale_settings=bot_alias_locale_settings,
                conversation_log_settings=conversation_log_settings,
                sentiment_analysis_settings=sentiment_analysis_settings,
                bot_id=bot_id,
            )
        )

    def list_bot_aliases(self) -> str:
        bot_id = self._get_param("botId")
        max_results = self._get_param("maxResults")
        bot_alias_summaries, bot_id = self.lexv2models_backend.list_bot_aliases(
            bot_id=bot_id, max_results=max_results
        )
        return json.dumps(
            dict(
                botAliasSummaries=bot_alias_summaries,
                nextToken=None,
                botId=bot_id,
            )
        )

    def delete_bot_alias(self) -> str:
        bot_alias_id = self._get_param("botAliasId")
        bot_id = self._get_param("botId")
        skip_resource_in_use_check = self._get_param("skipResourceInUseCheck")
        bot_alias_id, bot_id, bot_alias_status = (
            self.lexv2models_backend.delete_bot_alias(
                bot_alias_id=bot_alias_id,
                bot_id=bot_id,
                skip_resource_in_use_check=skip_resource_in_use_check,
            )
        )
        return json.dumps(
            dict(botAliasId=bot_alias_id, botId=bot_id, botAliasStatus=bot_alias_status)
        )

    def create_resource_policy(self) -> str:
        resource_arn = self._get_param("resourceArn")
        policy = self._get_param("policy")
        resource_arn, revision_id = self.lexv2models_backend.create_resource_policy(
            resource_arn=resource_arn,
            policy=policy,
        )
        return json.dumps(dict(resourceArn=resource_arn, revisionId=revision_id))

    def describe_resource_policy(self) -> str:
        resource_arn = self._get_param("resourceArn")
        resource_arn, policy, revision_id = (
            self.lexv2models_backend.describe_resource_policy(
                resource_arn=resource_arn,
            )
        )
        return json.dumps(
            dict(resourceArn=resource_arn, policy=policy, revisionId=revision_id)
        )

    def update_resource_policy(self) -> str:
        resource_arn = self._get_param("resourceArn")
        policy = self._get_param("policy")
        expected_revision_id = self._get_param("expectedRevisionId")
        resource_arn, revision_id = self.lexv2models_backend.update_resource_policy(
            resource_arn=resource_arn,
            policy=policy,
            expected_revision_id=expected_revision_id,
        )
        return json.dumps(dict(resourceArn=resource_arn, revisionId=revision_id))

    def delete_resource_policy(self) -> str:
        resource_arn = self._get_param("resourceArn")
        expected_revision_id = self._get_param("expectedRevisionId")
        resource_arn, revision_id = self.lexv2models_backend.delete_resource_policy(
            resource_arn=resource_arn,
            expected_revision_id=expected_revision_id,
        )
        return json.dumps(dict(resourceArn=resource_arn, revisionId=revision_id))

    def tag_resource(self) -> str:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tags = self._get_param("tags")
        self.lexv2models_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps(dict())

    def untag_resource(self) -> str:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tag_keys = self.__dict__["data"]["tagKeys"]
        self.lexv2models_backend.untag_resource(
            resource_arn=resource_arn,
            tag_keys=tag_keys,
        )
        return json.dumps(dict())

    def list_tags_for_resource(self) -> str:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tags = self.lexv2models_backend.list_tags_for_resource(
            resource_arn=resource_arn,
        )
        return json.dumps(dict(tags=tags))



import json
from typing import Any, Dict, List
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import Inspector2Backend, inspector2_backends


class Inspector2Response(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="inspector2")

    @property
    def inspector2_backend(self) -> Inspector2Backend:
        return inspector2_backends[self.current_account][self.region]

    def create_filter(self) -> str:
        action = self._get_param("action")
        description = self._get_param("description")
        filter_criteria = self._get_param("filterCriteria")
        name = self._get_param("name")
        reason = self._get_param("reason")
        tags = self._get_param("tags")
        arn = self.inspector2_backend.create_filter(
            action=action,
            description=description,
            filter_criteria=filter_criteria,
            name=name,
            reason=reason,
            tags=tags,
        )
        return json.dumps(dict(arn=arn))

    def delete_filter(self) -> str:
        arn = self._get_param("arn")
        self.inspector2_backend.delete_filter(arn=arn)
        return json.dumps(dict(arn=arn))

    def list_filters(self) -> str:
        action = self._get_param("action")
        arns = self._get_param("arns")
        filters = self.inspector2_backend.list_filters(action=action, arns=arns)
        return json.dumps({"filters": [f.to_json() for f in filters]})

    def list_findings(self) -> str:
        filter_criteria = self._get_param("filterCriteria")
        max_results = self._get_param("maxResults")
        next_token = self._get_param("nextToken")
        sort_criteria = self._get_param("sortCriteria")
        findings = self.inspector2_backend.list_findings(
            filter_criteria=filter_criteria,
            max_results=max_results,
            next_token=next_token,
            sort_criteria=sort_criteria,
        )
        return json.dumps(dict(findings=findings))

    def list_delegated_admin_accounts(self) -> str:
        accounts = self.inspector2_backend.list_delegated_admin_accounts()
        return json.dumps(
            {
                "delegatedAdminAccounts": [
                    {"accountId": key, "status": val} for key, val in accounts.items()
                ]
            }
        )

    def enable_delegated_admin_account(self) -> str:
        account_id = self._get_param("delegatedAdminAccountId")
        self.inspector2_backend.enable_delegated_admin_account(account_id)
        return json.dumps({"delegatedAdminAccountId": account_id})

    def disable_delegated_admin_account(self) -> str:
        account_id = self._get_param("delegatedAdminAccountId")
        self.inspector2_backend.disable_delegated_admin_account(account_id)
        return json.dumps({"delegatedAdminAccountId": account_id})

    def describe_organization_configuration(self) -> str:
        config = self.inspector2_backend.describe_organization_configuration()
        return json.dumps(config)

    def update_organization_configuration(self) -> str:
        auto_enable = self._get_param("autoEnable")
        config = self.inspector2_backend.update_organization_configuration(auto_enable)
        return json.dumps(config)

    def enable(self) -> str:
        account_ids = self._get_param("accountIds")
        resource_types = self._get_param("resourceTypes")
        accounts = self.inspector2_backend.enable(account_ids, resource_types)
        failed: List[Dict[str, Any]] = []
        return json.dumps({"accounts": accounts, "failedAccounts": failed})

    def disable(self) -> str:
        account_ids = self._get_param("accountIds")
        resource_types = self._get_param("resourceTypes")
        accounts = self.inspector2_backend.disable(account_ids, resource_types)
        failed: List[Dict[str, Any]] = []
        return json.dumps({"accounts": accounts, "failedAccounts": failed})

    def batch_get_account_status(self) -> str:
        account_ids = self._get_param("accountIds")
        accounts = self.inspector2_backend.batch_get_account_status(account_ids)
        failed: List[Dict[str, Any]] = []
        return json.dumps({"accounts": accounts, "failedAccounts": failed})

    def list_members(self) -> str:
        members = self.inspector2_backend.list_members()
        return json.dumps({"members": [m.to_json() for m in members]})

    def associate_member(self) -> str:
        account_id = self._get_param("accountId")
        self.inspector2_backend.associate_member(account_id)
        return json.dumps({"accountId": account_id})

    def disassociate_member(self) -> str:
        account_id = self._get_param("accountId")
        self.inspector2_backend.disassociate_member(account_id)
        return json.dumps({"accountId": account_id})

    def get_member(self) -> str:
        account_id = self._get_param("accountId")
        member = self.inspector2_backend.get_member(account_id)
        return json.dumps({"member": member.to_json()})

    def list_tags_for_resource(self) -> str:
        arn = unquote(self.path.split("/tags/")[-1])
        tags = self.inspector2_backend.list_tags_for_resource(arn)
        return json.dumps({"tags": tags})

    def tag_resource(self) -> str:
        resource_arn = unquote(self.path.split("/tags/")[-1])
        tags = self._get_param("tags")
        self.inspector2_backend.tag_resource(resource_arn=resource_arn, tags=tags)
        return "{}"

    def untag_resource(self) -> str:
        resource_arn = unquote(self.path.split("/tags/")[-1])
        tag_keys = self.querystring.get("tagKeys")
        self.inspector2_backend.untag_resource(resource_arn, tag_keys)  # type: ignore
        return "{}"



"""Unit tests for lexv2models-supported APIs."""

import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_bot():
    client = boto3.client("lexv2-models", region_name="ap-southeast-1")
    resp = client.create_bot(
        botName="test_bot",
        description="test_bot_description",
        roleArn="arn:aws:iam::123456789012:role/lex-role",
        dataPrivacy={"childDirected": False},
        idleSessionTTLInSeconds=300,
        botTags={"test_key": "test_value"},
        testBotAliasTags={"test_key": "test_value"},
        botType="test_bot_type",
        botMembers=[
            {
                "botMemberId": "test_bot_member_id",
                "botMemberName": "string",
                "botMemberAliasId": "test_bot_member_alias_id",
                "botMemberAliasName": "string",
                "botMemberVersion": "string",
            },
        ],
    )

    assert resp.get("botId")
    assert resp["botName"] == "test_bot"
    assert resp["description"] == "test_bot_description"
    assert resp["roleArn"] == "arn:aws:iam::123456789012:role/lex-role"
    assert resp["dataPrivacy"] == {"childDirected": False}
    assert resp["idleSessionTTLInSeconds"] == 300
    assert resp["botStatus"] == "CREATING"
    assert resp.get("creationDateTime")
    assert resp["botTags"] == {"test_key": "test_value"}
    assert resp["testBotAliasTags"] == {"test_key": "test_value"}
    assert resp["botType"] == "test_bot_type"
    assert resp["botMembers"] == [
        {
            "botMemberId": "test_bot_member_id",
            "botMemberName": "string",
            "botMemberAliasId": "test_bot_member_alias_id",
            "botMemberAliasName": "string",
            "botMemberVersion": "string",
        }
    ]
    assert resp.get("creationDateTime")


@mock_aws
def test_describe_bot():
    client = boto3.client("lexv2-models", region_name="eu-west-1")
    bot = client.create_bot(
        botName="test_bot",
        description="test_bot_description",
        roleArn="arn:aws:iam::123456789012:role/lex-role",
        dataPrivacy={"childDirected": False},
        idleSessionTTLInSeconds=300,
        botTags={"test_key": "test_value"},
        testBotAliasTags={"test_key": "test_value"},
        botType="test_bot_type",
        botMembers=[
            {
                "botMemberId": "test_bot_member_id",
                "botMemberName": "string",
                "botMemberAliasId": "test_bot_member_alias_id",
                "botMemberAliasName": "string",
                "botMemberVersion": "string",
            },
        ],
    )

    resp = client.describe_bot(botId=bot["botId"])

    assert resp["botId"] == bot["botId"]
    assert resp["botName"] == "test_bot"
    assert resp["description"] == "test_bot_description"
    assert resp["roleArn"] == "arn:aws:iam::123456789012:role/lex-role"
    assert resp["dataPrivacy"] == {"childDirected": False}
    assert resp["idleSessionTTLInSeconds"] == 300
    assert resp["botStatus"] == "CREATING"
    assert resp.get("creationDateTime")
    assert resp.get("lastUpdatedDateTime")
    assert resp["botType"] == "test_bot_type"
    assert resp["botMembers"] == [
        {
            "botMemberId": "test_bot_member_id",
            "botMemberName": "string",
            "botMemberAliasId": "test_bot_member_alias_id",
            "botMemberAliasName": "string",
            "botMemberVersion": "string",
        }
    ]
    assert resp["failureReasons"] == []


@mock_aws
def test_update_bot():
    client = boto3.client("lexv2-models", region_name="ap-southeast-1")
    bot = client.create_bot(
        botName="test_bot",
        description="test_bot_description",
        roleArn="arn:aws:iam::123456789012:role/lex-role",
        dataPrivacy={"childDirected": False},
        idleSessionTTLInSeconds=300,
        botTags={"test_key": "test_value"},
        testBotAliasTags={"test_key": "test_value"},
        botType="test_bot_type",
        botMembers=[
            {
                "botMemberId": "test_bot_member_id",
                "botMemberName": "string",
                "botMemberAliasId": "test_bot_member_alias_id",
                "botMemberAliasName": "string",
                "botMemberVersion": "string",
            },
        ],
    )

    resp = client.update_bot(
        botId=bot["botId"],
        botName="test_bot_updated",
        description="test_bot_description_updated",
        roleArn="arn:aws:iam::123456789012:role/lex-role-updated",
        dataPrivacy={"childDirected": True},
        idleSessionTTLInSeconds=600,
        botType="test_bot_type_updated",
        botMembers=[
            {
                "botMemberId": "test_bot_member_id_updated",
                "botMemberName": "string",
                "botMemberAliasId": "test_bot_member_alias_id_updated",
                "botMemberAliasName": "string",
                "botMemberVersion": "string",
            },
        ],
    )

    assert resp["botId"] == bot["botId"]
    assert resp["botName"] == "test_bot_updated"
    assert resp["description"] == "test_bot_description_updated"
    assert resp["roleArn"] == "arn:aws:iam::123456789012:role/lex-role-updated"
    assert resp["dataPrivacy"] == {"childDirected": True}
    assert resp["idleSessionTTLInSeconds"] == 600
    assert resp["botStatus"] == "CREATING"
    assert resp.get("creationDateTime")
    assert resp.get("lastUpdatedDateTime")
    assert resp.get("lastUpdatedDateTime") != resp.get("creationDateTime")
    assert resp["botType"] == "test_bot_type_updated"
    assert resp["botMembers"] == [
        {
            "botMemberId": "test_bot_member_id_updated",
            "botMemberName": "string",
            "botMemberAliasId": "test_bot_member_alias_id_updated",
            "botMemberAliasName": "string",
            "botMemberVersion": "string",
        }
    ]


@mock_aws
def test_list_bots():
    client = boto3.client("lexv2-models", region_name="ap-southeast-1")
    bot = client.create_bot(
        botName="test_bot",
        description="test_bot_description",
        roleArn="arn:aws:iam::123456789012:role/lex-role",
        dataPrivacy={"childDirected": False},
        idleSessionTTLInSeconds=300,
        botTags={"test_key": "test_value"},
        testBotAliasTags={"test_key": "test_value"},
        botType="test_bot_type",
        botMembers=[
            {
                "botMemberId": "test_bot_member_id",
                "botMemberName": "string",
                "botMemberAliasId": "test_bot_member_alias_id",
                "botMemberAliasName": "string",
                "botMemberVersion": "string",
            },
        ],
    )

    resp = client.list_bots()["botSummaries"]
    assert len(resp) == 1
    assert resp[0]["botId"] == bot["botId"]
    assert resp[0]["botName"] == "test_bot"
    assert resp[0]["description"] == "test_bot_description"
    assert resp[0]["botStatus"] == "CREATING"
    assert resp[0]["latestBotVersion"] == 1
    assert resp[0]["lastUpdatedDateTime"]
    assert resp[0]["botType"] == "test_bot_type"

    client.delete_bot(botId=bot["botId"], skipResourceInUseCheck=True)
    resp = client.list_bots()["botSummaries"]
    assert len(resp) == 0


@mock_aws
def test_bot_alias():
    client = boto3.client("lexv2-models", region_name="ap-southeast-1")
    resp = client.create_bot_alias(
        botAliasName="test_bot_alias",
        description="test_bot_alias_description",
        botVersion="1",
        botAliasLocaleSettings={"en-US": {"enabled": True}},
        conversationLogSettings={
            "textLogSettings": [
                {
                    "enabled": True,
                    "destination": {
                        "cloudWatch": {
                            "cloudWatchLogGroupArn": "test_log_group_arn",
                            "logPrefix": "test_log_prefix",
                        }
                    },
                }
            ]
        },
        sentimentAnalysisSettings={"detectSentiment": True},
        botId="test_bot_id",
        tags={"test_key": "test_value"},
    )

    assert resp.get("botAliasId")
    assert resp["botAliasName"] == "test_bot_alias"
    assert resp["description"] == "test_bot_alias_description"
    assert resp["botVersion"] == "1"
    assert resp["botAliasLocaleSettings"] == {"en-US": {"enabled": True}}
    assert resp["conversationLogSettings"] == {
        "textLogSettings": [
            {
                "enabled": True,
                "destination": {
                    "cloudWatch": {
                        "cloudWatchLogGroupArn": "test_log_group_arn",
                        "logPrefix": "test_log_prefix",
                    }
                },
            }
        ]
    }
    assert resp["sentimentAnalysisSettings"] == {"detectSentiment": True}
    assert resp["botAliasStatus"] == "CREATING"
    assert resp["botId"] == "test_bot_id"
    assert resp.get("creationDateTime")
    assert resp["tags"] == {"test_key": "test_value"}

    desc_resp = client.describe_bot_alias(
        botAliasId=resp["botAliasId"], botId="test_bot_id"
    )

    assert desc_resp["botAliasId"] == resp["botAliasId"]
    assert desc_resp["botAliasName"] == "test_bot_alias"
    assert desc_resp["description"] == "test_bot_alias_description"
    assert desc_resp["botVersion"] == "1"
    assert desc_resp["botAliasLocaleSettings"] == {"en-US": {"enabled": True}}
    assert desc_resp["conversationLogSettings"] == {
        "textLogSettings": [
            {
                "enabled": True,
                "destination": {
                    "cloudWatch": {
                        "cloudWatchLogGroupArn": "test_log_group_arn",
                        "logPrefix": "test_log_prefix",
                    }
                },
            }
        ]
    }
    assert desc_resp["sentimentAnalysisSettings"] == {"detectSentiment": True}
    assert desc_resp["botAliasHistoryEvents"] == []
    assert desc_resp["botAliasStatus"] == "CREATING"
    assert desc_resp["botId"] == "test_bot_id"
    assert desc_resp.get("creationDateTime")
    assert desc_resp.get("lastUpdatedDateTime")
    assert desc_resp["parentBotNetworks"] == []

    list_resp = client.list_bot_aliases(botId="test_bot_id")["botAliasSummaries"]

    assert len(list_resp) == 1
    assert list_resp[0]["botAliasId"] == resp["botAliasId"]
    assert list_resp[0]["botAliasName"] == "test_bot_alias"
    assert list_resp[0]["description"] == "test_bot_alias_description"
    assert list_resp[0]["botVersion"] == "1"
    assert list_resp[0]["botAliasStatus"] == "CREATING"
    assert list_resp[0]["creationDateTime"] == resp["creationDateTime"]

    update_resp = client.update_bot_alias(
        botAliasId=resp["botAliasId"],
        botId="test_bot_id",
        botAliasName="test_bot_alias_updated",
        description="test_bot_alias_description_updated",
        botVersion="1",
        botAliasLocaleSettings={"en-US": {"enabled": True}},
        conversationLogSettings={
            "textLogSettings": [
                {
                    "enabled": True,
                    "destination": {
                        "cloudWatch": {
                            "cloudWatchLogGroupArn": "test_log_group_arn_updated",
                            "logPrefix": "test_log_prefix_updated",
                        }
                    },
                }
            ]
        },
        sentimentAnalysisSettings={"detectSentiment": False},
    )
    assert update_resp["botAliasId"] == resp["botAliasId"]
    assert update_resp["botAliasName"] == "test_bot_alias_updated"
    assert update_resp["description"] == "test_bot_alias_description_updated"
    assert update_resp["botVersion"] == "1"
    assert update_resp["botAliasLocaleSettings"] == {"en-US": {"enabled": True}}
    assert update_resp["conversationLogSettings"] == {
        "textLogSettings": [
            {
                "enabled": True,
                "destination": {
                    "cloudWatch": {
                        "cloudWatchLogGroupArn": "test_log_group_arn_updated",
                        "logPrefix": "test_log_prefix_updated",
                    }
                },
            }
        ]
    }
    assert update_resp["sentimentAnalysisSettings"] == {"detectSentiment": False}
    assert update_resp["botAliasStatus"] == "CREATING"
    assert update_resp["botId"] == "test_bot_id"
    assert update_resp.get("lastUpdatedDateTime") != resp.get("creationDateTime")

    client.delete_bot_alias(botAliasId=resp["botAliasId"], botId="test_bot_id")
    list_resp = client.list_bot_aliases(botId="test_bot_id")["botAliasSummaries"]
    assert len(list_resp) == 0


@mock_aws
def test_resource_policy():
    client = boto3.client("lexv2-models", region_name="eu-west-1")
    resp = client.create_resource_policy(
        resourceArn="test_resource_arn",
        policy="test_resource_policy",
    )
    assert resp["resourceArn"] == "test_resource_arn"
    assert resp.get("revisionId")

    desc_resp = client.describe_resource_policy(
        resourceArn="test_resource_arn",
    )
    assert desc_resp["resourceArn"] == "test_resource_arn"
    assert desc_resp["policy"] == "test_resource_policy"
    assert desc_resp.get("revisionId")

    update_resp = client.update_resource_policy(
        resourceArn="test_resource_arn",
        policy="test_resource_policy_updated",
        expectedRevisionId=resp["revisionId"],
    )

    assert update_resp["resourceArn"] == "test_resource_arn"
    assert update_resp["revisionId"] != resp["revisionId"]

    delete_resp = client.delete_resource_policy(
        resourceArn="test_resource_arn",
        expectedRevisionId=update_resp["revisionId"],
    )
    assert delete_resp["resourceArn"] == "test_resource_arn"


@mock_aws
def test_tag_resource():
    sts = boto3.client("sts")
    account_id = sts.get_caller_identity()["Account"]
    region_name = "eu-west-1"

    client = boto3.client("lexv2-models", region_name="eu-west-1")
    cr_resp = client.create_bot(
        botName="test_bot",
        description="test_bot_description",
        roleArn="arn:aws:iam::123456789012:role/lex-role",
        dataPrivacy={"childDirected": False},
        idleSessionTTLInSeconds=300,
        botTags={"test_key": "test_value"},
        testBotAliasTags={"test_key": "test_value"},
        botType="test_bot_type",
        botMembers=[
            {
                "botMemberId": "test_bot_member_id",
                "botMemberName": "string",
                "botMemberAliasId": "test_bot_member_alias_id",
                "botMemberAliasName": "string",
                "botMemberVersion": "string",
            },
        ],
    )

    arn = f"arn:aws:lex:{region_name}:{account_id}:bot/{cr_resp['botId']}"

    resp = client.list_tags_for_resource(resourceARN=arn)
    assert resp["tags"] == {"test_key": "test_value"}

    client.tag_resource(resourceARN=arn, tags={"new_key": "new_value"})
    resp = client.list_tags_for_resource(resourceARN=arn)
    assert resp["tags"] == {"test_key": "test_value", "new_key": "new_value"}

    client.untag_resource(resourceARN=arn, tagKeys=["new_key"])
    resp = client.list_tags_for_resource(resourceARN=arn)
    assert resp["tags"] == {"test_key": "test_value"}



"""Unit tests for quicksight-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "GroupName": "mygroup",
            "Description": "my new fancy group",
        },
        {
            "GroupName": "users@munich",
            "Description": "all munich users",
        },
    ],
)
@mock_aws
def test_create_group(request_params):
    client = boto3.client("quicksight", region_name="us-west-2")
    resp = client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName"],
        Description=request_params["Description"],
    )

    assert "Group" in resp

    assert resp["Group"]["Arn"] == (
        f"arn:aws:quicksight:us-west-2:{ACCOUNT_ID}:group/default/{request_params['GroupName']}"
    )
    assert resp["Group"]["GroupName"] == request_params["GroupName"]
    assert resp["Group"]["Description"] == request_params["Description"]
    assert resp["Group"]["PrincipalId"] == f"{ACCOUNT_ID}"


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "GroupName": "mygroup",
            "Description": "my new fancy group",
        },
        {
            "GroupName": "users@munich",
            "Description": "all munich users",
        },
    ],
)
@mock_aws
def test_describe_group(request_params):
    client = boto3.client("quicksight", region_name="us-west-2")
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName"],
        Description=request_params["Description"],
    )

    resp = client.describe_group(
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert "Group" in resp

    assert resp["Group"]["Arn"] == (
        f"arn:aws:quicksight:us-west-2:{ACCOUNT_ID}:group/default/{request_params['GroupName']}"
    )
    assert resp["Group"]["GroupName"] == request_params["GroupName"]
    assert resp["Group"]["Description"] == request_params["Description"]
    assert resp["Group"]["PrincipalId"] == f"{ACCOUNT_ID}"


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "GroupName": "mygroup",
            "Description": "my new fancy group",
        },
        {
            "GroupName": "users@munich",
            "Description": "all munich users",
        },
    ],
)
@mock_aws
def test_update_group(request_params):
    client = boto3.client("quicksight", region_name="us-west-2")
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName"],
        Description="desc1",
    )

    resp = client.update_group(
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Description="desc2",
    )
    assert resp["Group"]["Description"] == "desc2"

    resp = client.describe_group(
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    assert "Group" in resp
    assert resp["Group"]["Arn"] == (
        f"arn:aws:quicksight:us-west-2:{ACCOUNT_ID}:group/default/{request_params['GroupName']}"
    )
    assert resp["Group"]["GroupName"] == request_params["GroupName"]
    assert resp["Group"]["Description"] == "desc2"
    assert resp["Group"]["PrincipalId"] == f"{ACCOUNT_ID}"


@pytest.mark.parametrize(
    "request_params",
    [
        {
            "GroupName": "mygroup",
            "Description": "my new fancy group",
        },
        {
            "GroupName": "users@munich",
            "Description": "all munich users",
        },
    ],
)
@mock_aws
def test_delete_group(request_params):
    client = boto3.client("quicksight", region_name="us-east-2")
    client.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName=request_params["GroupName"],
        Description=request_params["Description"],
    )

    client.delete_group(
        GroupName=request_params["GroupName"],
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
    )

    with pytest.raises(ClientError) as exc:
        client.describe_group(
            GroupName=request_params["GroupName"],
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_groups__initial():
    client = boto3.client("quicksight", region_name="us-east-2")
    resp = client.list_groups(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert resp["GroupList"] == []
    assert resp["Status"] == 200


@mock_aws
def test_list_groups():
    client = boto3.client("quicksight", region_name="us-east-1")
    for i in range(4):
        group_name = f"group{i}" if i < 2 else f"group{i}@test"
        client.create_group(
            AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName=group_name
        )

    resp = client.list_groups(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert len(resp["GroupList"]) == 4
    assert resp["Status"] == 200

    assert {
        "Arn": f"arn:aws:quicksight:us-east-1:{ACCOUNT_ID}:group/default/group0",
        "GroupName": "group0",
        "PrincipalId": ACCOUNT_ID,
    } in resp["GroupList"]

    assert {
        "Arn": f"arn:aws:quicksight:us-east-1:{ACCOUNT_ID}:group/default/group3@test",
        "GroupName": "group3@test",
        "PrincipalId": ACCOUNT_ID,
    } in resp["GroupList"]


@mock_aws
def test_list_groups__paginated():
    client = boto3.client("quicksight", region_name="us-east-1")
    for i in range(125):
        client.create_group(
            AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName=f"group{i}"
        )
    # default pagesize is 100
    page1 = client.list_groups(AwsAccountId=ACCOUNT_ID, Namespace="default")
    assert len(page1["GroupList"]) == 100
    assert "NextToken" in page1

    # We can ask for a smaller pagesize
    page2 = client.list_groups(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=15,
        NextToken=page1["NextToken"],
    )
    assert len(page2["GroupList"]) == 15
    assert "NextToken" in page2

    # We could request all of them in one go
    all_users = client.list_groups(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        MaxResults=1000,
    )
    length = len(all_users["GroupList"])
    # We don't know exactly how much workspaces there are, because we are running multiple tests at the same time
    assert length >= 125


@mock_aws
def test_search_groups():
    client = boto3.client("quicksight", region_name="us-east-1")
    for i in range(4):
        group_name = f"group{i}" if i < 2 else f"test{i}@test"
        client.create_group(
            AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName=group_name
        )

    resp = client.search_groups(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Filters=[
            {"Operator": "StringEquals", "Name": "GROUP_NAME", "Value": "group1"},
        ],
    )

    assert len(resp["GroupList"]) == 1
    assert resp["Status"] == 200

    assert {
        "Arn": f"arn:aws:quicksight:us-east-1:{ACCOUNT_ID}:group/default/group1",
        "GroupName": "group1",
        "PrincipalId": ACCOUNT_ID,
    } in resp["GroupList"]


@mock_aws
def test_search_groups__paginated():
    client = boto3.client("quicksight", region_name="us-east-1")
    for i in range(250):
        group_name = f"group{i}" if i % 2 else f"test{i}@test"
        client.create_group(
            AwsAccountId=ACCOUNT_ID, Namespace="default", GroupName=group_name
        )

    # default pagesize is 100
    page1 = client.search_groups(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Filters=[
            {"Operator": "StartsWith", "Name": "GROUP_NAME", "Value": "group"},
        ],
    )
    assert len(page1["GroupList"]) == 100
    assert "NextToken" in page1

    # We can ask for a smaller pagesize
    page2 = client.search_groups(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Filters=[
            {"Operator": "StartsWith", "Name": "GROUP_NAME", "Value": "group"},
        ],
        MaxResults=15,
        NextToken=page1["NextToken"],
    )
    assert len(page2["GroupList"]) == 15
    assert "NextToken" in page2

    # We could request all of them in one go
    all_users = client.search_groups(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        Filters=[
            {"Operator": "StartsWith", "Name": "GROUP_NAME", "Value": "group"},
        ],
        MaxResults=1000,
    )
    length = len(all_users["GroupList"])
    # We don't know exactly how much workspaces there are, because we are running multiple tests at the same time
    assert length >= 125


@mock_aws
def test_list_groups__diff_account_region():
    ACCOUNT_ID_2 = "998877665544"
    client_us = boto3.client("quicksight", region_name="us-east-2")
    client_eu = boto3.client("quicksight", region_name="eu-west-1")
    resp = client_us.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName="group_us_1",
        Description="Group in US Account 1",
    )
    resp = client_us.create_group(
        AwsAccountId=ACCOUNT_ID_2,
        Namespace="default",
        GroupName="group_us_2",
        Description="Group in US Account 2",
    )
    resp = client_eu.create_group(
        AwsAccountId=ACCOUNT_ID,
        Namespace="default",
        GroupName="group_eu_1",
        Description="Group in EU Account 1",
    )

    # Return Account 1, Region US
    resp = client_us.list_groups(AwsAccountId=ACCOUNT_ID, Namespace="default")
    assert len(resp["GroupList"]) == 1
    assert resp["Status"] == 200

    resp["GroupList"][0].pop("PrincipalId")

    assert resp["GroupList"][0] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID}:group/default/group_us_1",
        "GroupName": "group_us_1",
        "Description": "Group in US Account 1",
    }

    # Return Account 2, Region US
    resp = client_us.list_groups(AwsAccountId=ACCOUNT_ID_2, Namespace="default")

    assert len(resp["GroupList"]) == 1
    assert resp["Status"] == 200

    resp["GroupList"][0].pop("PrincipalId")

    assert resp["GroupList"][0] == {
        "Arn": f"arn:aws:quicksight:us-east-2:{ACCOUNT_ID_2}:group/default/group_us_2",
        "GroupName": "group_us_2",
        "Description": "Group in US Account 2",
    }

    # Return Account 1, Region EU
    resp = client_eu.list_groups(AwsAccountId=ACCOUNT_ID, Namespace="default")

    assert len(resp["GroupList"]) == 1
    assert resp["Status"] == 200

    resp["GroupList"][0].pop("PrincipalId")

    assert resp["GroupList"][0] == {
        "Arn": f"arn:aws:quicksight:eu-west-1:{ACCOUNT_ID}:group/default/group_eu_1",
        "GroupName": "group_eu_1",
        "Description": "Group in EU Account 1",
    }


@mock_aws
def test_search_groups__check_exceptions():
    client = boto3.client("quicksight", region_name="us-east-1")
    # Just do an exception test. No need to create a group first.

    with pytest.raises(ClientError) as exc:
        client.search_groups(
            AwsAccountId=ACCOUNT_ID,
            Namespace="default",
            Filters=[
                {
                    "Operator": "StringEquals",
                    "Name": "GROUP_DESCRIPTION",
                    "Value": "My Group 1",
                },
            ],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationException"



import json
import time
from unittest import SkipTest
from uuid import uuid4

import boto3
import pytest
import requests
from botocore.exceptions import ClientError

from moto import mock_aws, settings
from tests.test_awslambda.utilities import wait_for_log_msg

from ..markers import requires_docker
from .fixtures.custom_lambda import get_template, get_template_for_unknown_lambda


def get_lambda_code():
    return f"""
def lambda_handler(event, context):
    # Need to print this, one of the tests verifies the correct input
    print(event)
    response = dict()
    response["Status"] = "SUCCESS"
    response["StackId"] = event["StackId"]
    response["RequestId"] = event["RequestId"]
    response["LogicalResourceId"] = event["LogicalResourceId"]
    response["PhysicalResourceId"] = "CustomResource{str(uuid4())[0:6]}"
    response_data = dict()
    response_data["info_value"] = "special value"
    if event["RequestType"] == "Create":
        response["Data"] = response_data
    import cfnresponse
    cfnresponse.send(event, context, cfnresponse.SUCCESS, response_data)
"""


@mock_aws
def test_create_custom_lambda_resource():
    #########
    # Integration test using a Custom Resource
    # Create a Lambda
    # CF will call the Lambda
    # The Lambda should call CF, to indicate success (using the cfnresponse-module)
    # This HTTP request will include any outputs that are now stored against the stack
    # TEST: verify that this output is persisted
    ##########
    if not settings.TEST_SERVER_MODE:
        raise SkipTest(
            "Needs a standalone MotoServer, as cfnresponse needs to connect to something"
        )
    # Create cloudformation stack
    stack_name = f"stack{str(uuid4())[0:6]}"
    template_body = get_template(get_lambda_code())
    cf = boto3.client("cloudformation", region_name="us-east-1")
    cf.create_stack(
        StackName=stack_name,
        TemplateBody=json.dumps(template_body),
        Capabilities=["CAPABILITY_IAM"],
    )
    # Verify CloudWatch contains the correct logs
    log_group_name = get_log_group_name(cf, stack_name)
    success, logs = wait_for_log_msg(
        expected_msg="Status code: 200", log_group=log_group_name
    )
    assert success, f"Logs should indicate success: \n{logs}"

    # Verify the correct Output was returned
    outputs = get_outputs(cf, stack_name)
    assert len(outputs) == 1
    assert outputs[0]["OutputKey"] == "infokey"
    assert outputs[0]["OutputValue"] == "special value"


@mock_aws
@requires_docker
def test_create_custom_lambda_resource__verify_cfnresponse_failed():
    #########
    # Integration test using a Custom Resource
    # Create a Lambda
    # CF will call the Lambda
    # The Lambda should call CF --- this will fail, as we cannot make a HTTP request to the in-memory moto decorators
    # TEST: verify that the original event was send to the Lambda correctly
    # TEST: verify that a failure message appears in the CloudwatchLogs
    ##########
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Verify this fails if MotoServer is not running")
    # Create cloudformation stack
    stack_name = f"stack{str(uuid4())[0:6]}"
    template_body = get_template(get_lambda_code())
    cf = boto3.client("cloudformation", region_name="us-east-1")
    cf.create_stack(
        StackName=stack_name,
        TemplateBody=json.dumps(template_body),
        Capabilities=["CAPABILITY_IAM"],
    )
    # Verify CloudWatch contains the correct logs
    log_group_name = get_log_group_name(cf, stack_name)
    # urllib< 2 will emit the `failed executing http.request` message
    # urllib>=2 will emit the StatusCode=400 message
    execution_failed, logs = wait_for_log_msg(
        expected_msg=["failed executing http.request", "Status code: 400"],
        log_group=log_group_name,
    )
    assert execution_failed is True, logs

    printed_events = [
        line for line in logs if line.startswith("{'RequestType': 'Create'")
    ]
    assert len(printed_events) == 1
    original_event = json.loads(printed_events[0].replace("'", '"'))
    assert original_event["RequestType"] == "Create"
    assert "ServiceToken" in original_event  # Should equal Lambda ARN
    assert "ResponseURL" in original_event
    assert "StackId" in original_event
    assert "RequestId" in original_event  # type UUID
    assert original_event["LogicalResourceId"] == "CustomInfo"
    assert original_event["ResourceType"] == "Custom::Info"
    assert "ResourceProperties" in original_event
    assert (
        "ServiceToken" in original_event["ResourceProperties"]
    )  # Should equal Lambda ARN
    assert original_event["ResourceProperties"]["MyProperty"] == "stuff"


@mock_aws
def test_create_custom_lambda_resource__verify_manual_request():
    #########
    # Integration test using a Custom Resource
    # Create a Lambda
    # CF will call the Lambda
    # The Lambda should call CF --- this will fail, as we cannot make a HTTP request to the in-memory moto decorators
    # So we'll make this HTTP request manually
    # TEST: verify that the stack has a CREATE_IN_PROGRESS status before making the HTTP request
    # TEST: verify that the stack has a CREATE_COMPLETE status afterwards
    ##########
    if settings.TEST_SERVER_MODE:
        raise SkipTest(
            "Verify HTTP request can be made manually if MotoServer is not running"
        )
    # Create cloudformation stack
    stack_name = f"stack{str(uuid4())[0:6]}"
    template_body = get_template(get_lambda_code())
    region_name = "eu-north-1"
    cf = boto3.client("cloudformation", region_name=region_name)
    stack = cf.create_stack(
        StackName=stack_name,
        TemplateBody=json.dumps(template_body),
        Capabilities=["CAPABILITY_IAM"],
    )
    stack_id = stack["StackId"]
    stack = cf.describe_stacks(StackName=stack_id)["Stacks"][0]
    assert "Outputs" not in stack
    assert stack["StackStatus"] == "CREATE_IN_PROGRESS"

    callback_url = f"http://cloudformation.{region_name}.amazonaws.com/cloudformation_{region_name}/cfnresponse?stack={stack_id}"
    data = {
        "Status": "SUCCESS",
        "StackId": stack_id,
        "LogicalResourceId": "CustomInfo",
        "Data": {"info_value": "resultfromthirdpartysystem"},
    }
    requests.post(callback_url, json=data)

    stack = cf.describe_stacks(StackName=stack_id)["Stacks"][0]
    assert stack["StackStatus"] == "CREATE_COMPLETE"
    assert stack["Outputs"] == [
        {
            "OutputKey": "infokey",
            "OutputValue": "resultfromthirdpartysystem",
            "Description": "A very important value",
        },
    ]

    # AWSlambda will not have logged anything
    log_group_name = get_log_group_name(cf=cf, stack_name=stack_name)
    success, logs = wait_for_log_msg(
        expected_msg="Status code: 200", log_group=log_group_name, wait_time=5
    )
    assert success is False
    assert len(logs) == 0


@mock_aws
def test_create_custom_lambda_resource__unknown_arn():
    # Try to create a Lambda with an unknown ARN
    # Verify that this fails in a predictable manner
    cf = boto3.client("cloudformation", region_name="eu-north-1")
    with pytest.raises(ClientError) as exc:
        cf.create_stack(
            StackName=f"stack{str(uuid4())[0:6]}",
            TemplateBody=json.dumps(get_template_for_unknown_lambda()),
            Capabilities=["CAPABILITY_IAM"],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ValidationError"
    assert (
        err["Message"]
        == "Template error: instance of Fn::GetAtt references undefined resource InfoFunction"
    )


def get_log_group_name(cf, stack_name):
    resources = cf.describe_stack_resources(StackName=stack_name)["StackResources"]
    start = time.time()
    while (time.time() - start) < 5:
        fns = [
            r
            for r in resources
            if r["ResourceType"] == "AWS::Lambda::Function"
            and "PhysicalResourceId" in r
        ]
        if not fns:
            time.sleep(1)
            resources = cf.describe_stack_resources(StackName=stack_name)[
                "StackResources"
            ]
            continue

        fn = fns[0]
        resource_id = fn["PhysicalResourceId"]
        return f"/aws/lambda/{resource_id}"
    raise Exception("Could not find log group name in time")


def get_outputs(cf, stack_name):
    stack = cf.describe_stacks(StackName=stack_name)["Stacks"][0]
    start = time.time()
    while (time.time() - start) < 5:
        status = stack["StackStatus"]
        if status != "CREATE_COMPLETE":
            time.sleep(1)
            stack = cf.describe_stacks(StackName=stack_name)["Stacks"][0]
            continue

        outputs = stack["Outputs"]
        return outputs



from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ActivityResource,
    Resource,
    ResourceCondition,
    ServiceResource,
)
from moto.stepfunctions.parser.asl.static_analyser.static_analyser import StaticAnalyser


class ExpressStaticAnalyser(StaticAnalyser):
    def visitResource_decl(self, ctx: ASLParser.Resource_declContext) -> None:
        # TODO add resource path to the error messages.

        resource_str: str = ctx.keyword_or_string().getText()[1:-1]
        resource = Resource.from_resource_arn(resource_str)

        if isinstance(resource, ActivityResource):
            raise ValueError(
                "Invalid State Machine Definition: 'SCHEMA_VALIDATION_FAILED: "
                "Express state machine does not support Activity ARN'"
            )

        if isinstance(resource, ServiceResource):
            if resource.condition == ResourceCondition.WaitForTaskToken:
                raise ValueError(
                    "Invalid State Machine Definition: 'SCHEMA_VALIDATION_FAILED: "
                    "Express state machine does not support '.sync' service integration."
                )
            if resource.condition is not None:
                raise ValueError(
                    "Invalid State Machine Definition: 'SCHEMA_VALIDATION_FAILED: "
                    f"Express state machine does not support .'{resource.condition}' service integration."
                )



"""BedrockBackend class with methods for supported APIs."""

import re
from datetime import datetime
from typing import Any, Dict, List, Optional

from moto.bedrock.exceptions import (
    ResourceInUseException,
    ResourceNotFoundException,
    TooManyTagsException,
    ValidationException,
)
from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition


class ModelCustomizationJob(BaseModel):
    def __init__(
        self,
        job_name: str,
        custom_model_name: str,
        role_arn: str,
        base_model_identifier: str,
        training_data_config: Dict[str, str],
        output_data_config: Dict[str, str],
        hyper_parameters: Dict[str, str],
        region_name: str,
        account_id: str,
        client_request_token: Optional[str],
        customization_type: Optional[str],
        custom_model_kms_key_id: Optional[str],
        job_tags: Optional[List[Dict[str, str]]],
        custom_model_tags: Optional[List[Dict[str, str]]],
        validation_data_config: Optional[Dict[str, Any]],
        vpc_config: Optional[Dict[str, Any]],
    ):
        self.job_name = job_name
        self.custom_model_name = custom_model_name
        self.role_arn = role_arn
        self.client_request_token = client_request_token
        self.base_model_identifier = base_model_identifier
        self.customization_type = customization_type
        self.custom_model_kms_key_id = custom_model_kms_key_id
        self.job_tags = job_tags
        self.custom_model_tags = custom_model_tags
        if "s3Uri" not in training_data_config or not re.match(
            r"s3://.*", training_data_config["s3Uri"]
        ):
            raise ValidationException(
                "Validation error detected: "
                f"Value '{training_data_config}' at 'training_data_config' failed to satisfy constraint: "
                "Member must satisfy regular expression pattern: "
                "s3://.*"
            )
        self.training_data_config = training_data_config
        if validation_data_config:
            if "validators" in validation_data_config:
                for validator in validation_data_config["validators"]:
                    if not re.match(r"s3://.*", validator["s3Uri"]):
                        raise ValidationException(
                            "Validation error detected: "
                            f"Value '{validator}' at 'validation_data_config' failed to satisfy constraint: "
                            "Member must satisfy regular expression pattern: "
                            "s3://.*"
                        )
        self.validation_data_config = validation_data_config
        if "s3Uri" not in output_data_config or not re.match(
            r"s3://.*", output_data_config["s3Uri"]
        ):
            raise ValidationException(
                "Validation error detected: "
                f"Value '{output_data_config}' at 'output_data_config' failed to satisfy constraint: "
                "Member must satisfy regular expression pattern: "
                "s3://.*"
            )
        self.output_data_config = output_data_config
        self.hyper_parameters = hyper_parameters
        self.vpc_config = vpc_config
        self.region_name = region_name
        self.account_id = account_id
        self.job_arn = f"arn:{get_partition(self.region_name)}:bedrock:{self.region_name}:{self.account_id}:model-customization-job/{self.job_name}"
        self.output_model_name = f"{self.custom_model_name}-{self.job_name}"
        self.output_model_arn = f"arn:{get_partition(self.region_name)}:bedrock:{self.region_name}:{self.account_id}:custom-model/{self.output_model_name}"
        self.status = "InProgress"
        self.failure_message = "Failure Message"
        self.creation_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.last_modified_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.end_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.base_model_arn = f"arn:{get_partition(self.region_name)}:bedrock:{self.region_name}::foundation-model/{self.base_model_identifier}"
        self.output_model_kms_key_arn = f"arn:{get_partition(self.region_name)}:kms:{self.region_name}:{self.account_id}:key/{self.output_model_name}-kms-key"
        self.training_metrics = {"trainingLoss": 0.0}  # hard coded
        self.validation_metrics = [{"validationLoss": 0.0}]  # hard coded

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "baseModelArn": self.base_model_arn,
            "clientRequestToken": self.client_request_token,
            "creationTime": self.creation_time,
            "customizationType": self.customization_type,
            "endTime": self.end_time,
            "failureMessage": self.failure_message,
            "hyperParameters": self.hyper_parameters,
            "jobArn": self.job_arn,
            "jobName": self.job_name,
            "lastModifiedTime": self.last_modified_time,
            "outputDataConfig": self.output_data_config,
            "outputModelArn": self.output_model_arn,
            "outputModelKmsKeyArn": self.output_model_kms_key_arn,
            "outputModelName": self.output_model_name,
            "roleArn": self.role_arn,
            "status": self.status,
            "trainingDataConfig": self.training_data_config,
            "trainingMetrics": self.training_metrics,
            "validationDataConfig": self.validation_data_config,
            "validationMetrics": self.validation_metrics,
            "vpcConfig": self.vpc_config,
        }
        return {k: v for k, v in dct.items() if v}


class CustomModel(BaseModel):
    def __init__(
        self,
        model_name: str,
        job_name: str,
        job_arn: str,
        base_model_arn: str,
        hyper_parameters: Dict[str, str],
        output_data_config: Dict[str, str],
        training_data_config: Dict[str, str],
        training_metrics: Dict[str, float],
        base_model_name: str,
        region_name: str,
        account_id: str,
        customization_type: Optional[str],
        model_kms_key_arn: Optional[str],
        validation_data_config: Optional[Dict[str, Any]],
        validation_metrics: Optional[List[Dict[str, float]]],
    ):
        self.model_name = model_name
        self.job_name = job_name
        self.job_arn = job_arn
        self.base_model_arn = base_model_arn
        self.customization_type = customization_type
        self.model_kms_key_arn = model_kms_key_arn
        self.hyper_parameters = hyper_parameters
        self.training_data_config = training_data_config
        self.validation_data_config = validation_data_config
        self.output_data_config = output_data_config
        self.training_metrics = training_metrics
        self.validation_metrics = validation_metrics
        self.region_name = region_name
        self.account_id = account_id
        self.model_arn = f"arn:{get_partition(self.region_name)}:bedrock:{self.region_name}:{self.account_id}:custom-model/{self.model_name}"
        self.creation_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        self.base_model_name = base_model_name

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "baseModelArn": self.base_model_arn,
            "creationTime": self.creation_time,
            "customizationType": self.customization_type,
            "hyperParameters": self.hyper_parameters,
            "jobArn": self.job_arn,
            "jobName": self.job_name,
            "modelArn": self.model_arn,
            "modelKmsKeyArn": self.model_kms_key_arn,
            "modelName": self.model_name,
            "outputDataConfig": self.output_data_config,
            "trainingDataConfig": self.training_data_config,
            "trainingMetrics": self.training_metrics,
            "validationDataConfig": self.validation_data_config,
            "validationMetrics": self.validation_metrics,
        }
        return {k: v for k, v in dct.items() if v}


class model_invocation_logging_configuration(BaseModel):
    def __init__(self, logging_config: Dict[str, Any]) -> None:
        self.logging_config = logging_config


class BedrockBackend(BaseBackend):
    """Implementation of Bedrock APIs."""

    PAGINATION_MODEL = {
        "list_model_customization_jobs": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "job_arn",
        },
        "list_custom_models": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "model_arn",
        },
    }

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.model_customization_jobs: Dict[str, ModelCustomizationJob] = {}
        self.custom_models: Dict[str, CustomModel] = {}
        self.model_invocation_logging_configuration: Optional[
            model_invocation_logging_configuration
        ] = None
        self.tagger = TaggingService()

    def _list_arns(self) -> List[str]:
        return [job.job_arn for job in self.model_customization_jobs.values()] + [
            model.model_arn for model in self.custom_models.values()
        ]

    def create_model_customization_job(
        self,
        job_name: str,
        custom_model_name: str,
        role_arn: str,
        base_model_identifier: str,
        training_data_config: Dict[str, Any],
        output_data_config: Dict[str, str],
        hyper_parameters: Dict[str, str],
        client_request_token: Optional[str],
        customization_type: Optional[str],
        custom_model_kms_key_id: Optional[str],
        job_tags: Optional[List[Dict[str, str]]],
        custom_model_tags: Optional[List[Dict[str, str]]],
        validation_data_config: Optional[Dict[str, Any]],
        vpc_config: Optional[Dict[str, Any]],
    ) -> str:
        if job_name in self.model_customization_jobs.keys():
            raise ResourceInUseException(
                f"Model customization job {job_name} already exists"
            )
        if custom_model_name in self.custom_models.keys():
            raise ResourceInUseException(
                f"Custom model {custom_model_name} already exists"
            )
        model_customization_job = ModelCustomizationJob(
            job_name,
            custom_model_name,
            role_arn,
            base_model_identifier,
            training_data_config,
            output_data_config,
            hyper_parameters,
            self.region_name,
            self.account_id,
            client_request_token,
            customization_type,
            custom_model_kms_key_id,
            job_tags,
            custom_model_tags,
            validation_data_config,
            vpc_config,
        )
        self.model_customization_jobs[job_name] = model_customization_job
        if job_tags:
            self.tag_resource(model_customization_job.job_arn, job_tags)
        # Create associated custom model
        custom_model = CustomModel(
            custom_model_name,
            job_name,
            model_customization_job.job_arn,
            model_customization_job.base_model_arn,
            model_customization_job.hyper_parameters,
            model_customization_job.output_data_config,
            model_customization_job.training_data_config,
            model_customization_job.training_metrics,
            model_customization_job.base_model_identifier,
            self.region_name,
            self.account_id,
            model_customization_job.customization_type,
            model_customization_job.output_model_kms_key_arn,
            model_customization_job.validation_data_config,
            model_customization_job.validation_metrics,
        )
        self.custom_models[custom_model_name] = custom_model
        if custom_model_tags:
            self.tag_resource(custom_model.model_arn, custom_model_tags)
        return model_customization_job.job_arn

    def get_model_customization_job(self, job_identifier: str) -> ModelCustomizationJob:
        if job_identifier not in self.model_customization_jobs:
            raise ResourceNotFoundException(
                f"Model customization job {job_identifier} not found"
            )
        else:
            return self.model_customization_jobs[job_identifier]

    def stop_model_customization_job(self, job_identifier: str) -> None:
        if job_identifier in self.model_customization_jobs:
            self.model_customization_jobs[job_identifier].status = "Stopped"
        else:
            raise ResourceNotFoundException(
                f"Model customization job {job_identifier} not found"
            )
        return

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_model_customization_jobs(
        self,
        creation_time_after: Optional[datetime],
        creation_time_before: Optional[datetime],
        status_equals: Optional[str],
        name_contains: Optional[str],
        sort_by: Optional[str],
        sort_order: Optional[str],
    ) -> List[ModelCustomizationJob]:
        customization_jobs_fetched = list(self.model_customization_jobs.values())

        if name_contains is not None:
            customization_jobs_fetched = list(
                filter(
                    lambda x: name_contains in x.job_name,
                    customization_jobs_fetched,
                )
            )

        if creation_time_after is not None:
            customization_jobs_fetched = list(
                filter(
                    lambda x: x.creation_time > str(creation_time_after),
                    customization_jobs_fetched,
                )
            )

        if creation_time_before is not None:
            customization_jobs_fetched = list(
                filter(
                    lambda x: x.creation_time < str(creation_time_before),
                    customization_jobs_fetched,
                )
            )
        if status_equals is not None:
            customization_jobs_fetched = list(
                filter(
                    lambda x: x.status == status_equals,
                    customization_jobs_fetched,
                )
            )

        if sort_by is not None:
            if sort_by == "CreationTime":
                if sort_order is not None and sort_order == "Ascending":
                    customization_jobs_fetched = sorted(
                        customization_jobs_fetched, key=lambda x: x.creation_time
                    )
                elif sort_order is not None and sort_order == "Descending":
                    customization_jobs_fetched = sorted(
                        customization_jobs_fetched,
                        key=lambda x: x.creation_time,
                        reverse=True,
                    )
                else:
                    raise ValidationException(f"Invalid sort order: {sort_order}")
            else:
                raise ValidationException(f"Invalid sort by field: {sort_by}")

        return customization_jobs_fetched

    def get_model_invocation_logging_configuration(self) -> Optional[Dict[str, Any]]:
        if self.model_invocation_logging_configuration:
            return self.model_invocation_logging_configuration.logging_config
        else:
            return {}

    def put_model_invocation_logging_configuration(
        self, logging_config: Dict[str, Any]
    ) -> None:
        invocation_logging = model_invocation_logging_configuration(logging_config)
        self.model_invocation_logging_configuration = invocation_logging
        return

    def get_custom_model(self, model_identifier: str) -> CustomModel:
        if model_identifier[:3] == "arn":
            for model in self.custom_models.values():
                if model.model_arn == model_identifier:
                    return model
            raise ResourceNotFoundException(
                f"Custom model {model_identifier} not found"
            )
        elif model_identifier in self.custom_models:
            return self.custom_models[model_identifier]
        else:
            raise ResourceNotFoundException(
                f"Custom model {model_identifier} not found"
            )

    def delete_custom_model(self, model_identifier: str) -> None:
        if model_identifier in self.custom_models:
            del self.custom_models[model_identifier]
        else:
            raise ResourceNotFoundException(
                f"Custom model {model_identifier} not found"
            )
        return

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_custom_models(
        self,
        creation_time_before: Optional[datetime],
        creation_time_after: Optional[datetime],
        name_contains: Optional[str],
        base_model_arn_equals: Optional[str],
        foundation_model_arn_equals: Optional[str],
        sort_by: Optional[str],
        sort_order: Optional[str],
    ) -> List[CustomModel]:
        """
        The foundation_model_arn_equals-argument is not yet supported
        """
        custom_models_fetched = list(self.custom_models.values())

        if name_contains is not None:
            custom_models_fetched = list(
                filter(
                    lambda x: name_contains in x.job_name,
                    custom_models_fetched,
                )
            )

        if creation_time_after is not None:
            custom_models_fetched = list(
                filter(
                    lambda x: x.creation_time > str(creation_time_after),
                    custom_models_fetched,
                )
            )

        if creation_time_before is not None:
            custom_models_fetched = list(
                filter(
                    lambda x: x.creation_time < str(creation_time_before),
                    custom_models_fetched,
                )
            )
        if base_model_arn_equals is not None:
            custom_models_fetched = list(
                filter(
                    lambda x: x.base_model_arn == base_model_arn_equals,
                    custom_models_fetched,
                )
            )

        if sort_by is not None:
            if sort_by == "CreationTime":
                if sort_order is not None and sort_order == "Ascending":
                    custom_models_fetched = sorted(
                        custom_models_fetched, key=lambda x: x.creation_time
                    )
                elif sort_order is not None and sort_order == "Descending":
                    custom_models_fetched = sorted(
                        custom_models_fetched,
                        key=lambda x: x.creation_time,
                        reverse=True,
                    )
                else:
                    raise ValidationException(f"Invalid sort order: {sort_order}")
            else:
                raise ValidationException(f"Invalid sort by field: {sort_by}")
        return custom_models_fetched

    def tag_resource(self, resource_arn: str, tags: List[Dict[str, str]]) -> None:
        if resource_arn not in self._list_arns():
            raise ResourceNotFoundException(f"Resource {resource_arn} not found")
        fixed_tags = []
        if len(tags) + len(self.tagger.list_tags_for_resource(resource_arn)) > 50:
            raise TooManyTagsException(
                "Member must have length less than or equal to 50"
            )
        for tag_dict in tags:
            fixed_tags.append({"Key": tag_dict["key"], "Value": tag_dict["value"]})
        self.tagger.tag_resource(resource_arn, fixed_tags)
        return

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> None:
        if resource_arn not in self._list_arns():
            raise ResourceNotFoundException(f"Resource {resource_arn} not found")
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)
        return

    def list_tags_for_resource(self, resource_arn: str) -> List[Dict[str, str]]:
        if resource_arn not in self._list_arns():
            raise ResourceNotFoundException(f"Resource {resource_arn} not found")
        tags = self.tagger.list_tags_for_resource(resource_arn)
        fixed_tags = []
        for tag_dict in tags["Tags"]:
            fixed_tags.append({"key": tag_dict["Key"], "value": tag_dict["Value"]})
        return fixed_tags

    def delete_model_invocation_logging_configuration(self) -> None:
        if self.model_invocation_logging_configuration:
            self.model_invocation_logging_configuration.logging_config = {}
        return


bedrock_backends = BackendDict(BedrockBackend, "bedrock")



import json
from time import sleep
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from tests import allow_aws_request

from . import DEFAULT_REGION

ASSUME_ROLE_POLICY = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Effect": "Allow",
            "Principal": {"Service": "rds.amazonaws.com"},
            "Action": "sts:AssumeRole",
        }
    ],
}


@pytest.fixture(scope="module")
def mockaws():
    if allow_aws_request():
        yield
    else:
        with mock_aws():
            yield


@pytest.fixture(scope="module")
def secrets_arn():
    client = boto3.client("secretsmanager", DEFAULT_REGION)
    secret_name = f"moto-test-{str(uuid4())[0:6]}"

    try:
        secret = client.create_secret(Name=secret_name, SecretString="ss")
        yield secret["ARN"]
    finally:
        client.delete_secret(SecretId=secret_name)


@pytest.fixture(scope="module")
def vpc_id(mockaws):  # pylint: disable=redefined-outer-name
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    _vpc_id = None
    try:
        _vpc_id = ec2_client.create_vpc(CidrBlock="10.0.0.0/16")["Vpc"]["VpcId"]
        yield _vpc_id
    finally:
        ec2_client.delete_vpc(VpcId=_vpc_id)


@pytest.fixture(scope="module")
def subnet_id1(vpc_id):  # pylint: disable=redefined-outer-name
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    try:
        subnet = ec2_client.create_subnet(
            CidrBlock="10.0.1.0/24", VpcId=vpc_id, AvailabilityZone="us-east-1a"
        )
        subnet_id = subnet["Subnet"]["SubnetId"]
        yield subnet_id
    finally:
        ec2_client.delete_subnet(SubnetId=subnet_id)


@pytest.fixture(scope="module")
def subnet_id2(vpc_id):  # pylint: disable=redefined-outer-name
    ec2_client = boto3.client("ec2", region_name=DEFAULT_REGION)
    try:
        subnet = ec2_client.create_subnet(
            CidrBlock="10.0.2.0/24", VpcId=vpc_id, AvailabilityZone="us-east-1b"
        )
        subnet_id = subnet["Subnet"]["SubnetId"]
        yield subnet_id
    finally:
        ec2_client.delete_subnet(SubnetId=subnet_id)


@pytest.fixture(scope="module")
def role_arn(mockaws):  # pylint: disable=redefined-outer-name
    role_name = f"moto-test-{str(uuid4())[0:6]}"
    iam = boto3.client("iam", region_name=DEFAULT_REGION)
    try:
        role = iam.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(ASSUME_ROLE_POLICY),
        )["Role"]
        yield role["Arn"]
    finally:
        iam.delete_role(RoleName=role_name)


@pytest.fixture(scope="module")
def proxy_name(subnet_id1, subnet_id2, role_arn, secrets_arn):  # pylint: disable=redefined-outer-name
    name = f"moto-test-{str(uuid4())[0:6]}"
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    proxy_creation_succeeded = False
    try:
        resp = rds_client.create_db_proxy(
            DBProxyName=name,
            EngineFamily="POSTGRESQL",
            Auth=[
                {
                    # "UserName": "user1",
                    # "AuthScheme": "SECRETS",
                    "IAMAuth": "DISABLED",
                    "SecretArn": secrets_arn,
                }
            ],
            RoleArn=role_arn,
            VpcSubnetIds=[subnet_id1, subnet_id2],
        )
        status = resp["DBProxy"]["Status"]
        while status.lower() == "creating":
            sleep(5)
            status = rds_client.describe_db_proxies(DBProxyName=name)["DBProxies"][0][
                "Status"
            ]
        proxy_creation_succeeded = True

        yield name
    finally:
        if proxy_creation_succeeded:
            rds_client.delete_db_proxy(DBProxyName=name)
            deleted = False
            while not deleted:
                try:
                    rds_client.describe_db_proxies(DBProxyName=name)
                    sleep(10)
                except (ClientError, IndexError):
                    deleted = True


@pytest.fixture(scope="module")
def db_cluster_id(mockaws):  # pylint: disable=redefined-outer-name
    cluster_name = f"moto-test-{str(uuid4())[0:6]}"
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    cluster_creation_succeeded = False
    try:
        cluster_id = rds_client.create_db_cluster(
            DBClusterIdentifier=cluster_name,
            Engine="aurora-postgresql",
            MasterUsername="root",
            MasterUserPassword="hunter21",
        )["DBCluster"]["DBClusterIdentifier"]
        rds_client.get_waiter("db_cluster_available").wait(
            DBClusterIdentifier=cluster_id
        )
        cluster_creation_succeeded = True
        yield cluster_id
    finally:
        if cluster_creation_succeeded:
            rds_client.delete_db_cluster(
                DBClusterIdentifier=cluster_id, SkipFinalSnapshot=True
            )


def test_default_proxy_targets(account_id, proxy_name):  # pylint: disable=redefined-outer-name
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)
    resp = rds_client.describe_db_proxy_targets(DBProxyName=proxy_name)

    assert resp["Targets"] == []

    groups = rds_client.describe_db_proxy_target_groups(
        DBProxyName=proxy_name,
        TargetGroupName="default",
    )["TargetGroups"]
    assert len(groups) == 1
    groups[0].pop("CreatedDate")
    groups[0].pop("UpdatedDate")
    target_group_arn = groups[0].pop("TargetGroupArn")
    assert target_group_arn.startswith(
        f"arn:aws:rds:{DEFAULT_REGION}:{account_id}:target-group:prx-tg-"
    )  # 17 more chars (lowercase + digits)
    assert groups[0] == {
        "DBProxyName": proxy_name,
        "TargetGroupName": "default",
        "IsDefault": True,
        "Status": "available",
        "ConnectionPoolConfig": {
            "MaxConnectionsPercent": 100,
            "MaxIdleConnectionsPercent": 50,
            "ConnectionBorrowTimeout": 120,
            "SessionPinningFilters": [],
        },
    }


def test_register_db_proxy(account_id, proxy_name, db_cluster_id):  # pylint: disable=redefined-outer-name
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)

    resp = rds_client.register_db_proxy_targets(
        DBProxyName=proxy_name,
        DBClusterIdentifiers=[db_cluster_id],
    )["DBProxyTargets"]
    resp[0].pop("Endpoint", None)
    assert resp == [
        {
            "RdsResourceId": db_cluster_id,
            "Port": 5432,
            "Type": "TRACKED_CLUSTER",
            "TargetHealth": {"State": "REGISTERING"},
        }
    ]

    resp = rds_client.describe_db_proxy_targets(DBProxyName=proxy_name)["Targets"]
    resp[0].pop("Endpoint", None)
    resp[0].pop("TargetHealth", None)
    assert resp == [
        {"RdsResourceId": db_cluster_id, "Port": 5432, "Type": "TRACKED_CLUSTER"}
    ]

    rds_client.deregister_db_proxy_targets(
        DBProxyName=proxy_name,
        DBClusterIdentifiers=[db_cluster_id],
    )

    resp = rds_client.describe_db_proxy_targets(DBProxyName=proxy_name)
    assert resp["Targets"] == []


def test_modify_group(proxy_name):  # pylint: disable=redefined-outer-name
    rds_client = boto3.client("rds", region_name=DEFAULT_REGION)

    rds_client.modify_db_proxy_target_group(
        TargetGroupName="default",
        DBProxyName=proxy_name,
        ConnectionPoolConfig={
            "MaxConnectionsPercent": 27,
            "SessionPinningFilters": ["filter1"],
        },
    )

    group = rds_client.describe_db_proxy_target_groups(
        DBProxyName=proxy_name,
        TargetGroupName="default",
    )["TargetGroups"][0]
    assert group["ConnectionPoolConfig"]["MaxConnectionsPercent"] == 27
    assert group["ConnectionPoolConfig"]["MaxIdleConnectionsPercent"] == 13
    assert group["ConnectionPoolConfig"]["SessionPinningFilters"] == ["filter1"]
